#!/usr/bin/env python3.5

import  abc

import  sys
from    io              import  BytesIO as IO
from    pathlib         import  Path
import  cgi
import  json
import  urllib
import  ast
import  shutil
import  datetime
import  time
import  inspect
import  pprint

import  threading
import  platform
import  socket
import  psutil
import  os
import  multiprocessing
import  configparser
import  swiftclient
import  traceback
from    argparse            import  Namespace

import  pfmisc

# debugging utilities
import  pudb

from    pypx                import  repack

# pfstorage local dependencies
from    pfmisc._colors      import  Colors
from    pfmisc.debug        import  debug
from    pfmisc.C_snode      import  *
from    pfstate             import  S

def static_vars(**kwargs):
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate


class D(S):
    """
    A derived 'pfstate' class that keeps system state.
    """

    def __init__(self, arg, *args, **kwargs):
        """
        Constructor
        """

        proto = 'https' if arg['str_swiftPort'] == '443' else 'http'
        self.state_create(
        {
            "swift": {
                "auth_url":                 "%s://%s:%s/auth/v1.0" % \
                                            (proto, arg['str_swiftIP'], arg['str_swiftPort']),
                "username":                 arg['str_swiftLogin'],
                "key":                      "testing",
                "container_name":           "users",
                "auto_create_container":    True,
                "file_storage":             "swift.storage.SwiftStorage"
            }
        },
        *args, **kwargs
        )

        self.log    = self.dp.qprint

class PfStorage(metaclass = abc.ABCMeta):

    def state(self, *args):
        """
        Interact with the state object.

        If called with one arg, return the value at that location
        in the state tree representation.

        If called with two args, set the value at a given tree
        state representation.

        For example,

            key = state('/swift/key')

        returns the "/swift/key" while

            state('/swify/key', "a new value")

        sets "/swift/key" to "a new value"
        """

        if len(args) == 1:
            return self.S.T.cat(args[0])
        else:
            self.S.T.touch(args[0], args[1])

    def __init__(self, arg, *args, **kwargs):
        """
        The core constructor -- essentially this adds the main
        operational module objects to this class.
        """

        # pudb.set_trace()
        self.arg            = arg
        self.S              = D(arg, *args, **dict(kwargs, useGlobalState = True))
        self.dp             = pfmisc.debug(
                                verbosity   = self.state('/this/verbosity'),
                                within      = self.state('/this/name')
                            )
        self.log            = self.dp.qprint
        self.packer         = repack.Process(
                                repack.args_impedanceMatch(Namespace(**arg))
                            )
        # A generic member placeholder for an unspecified object --
        # In the case of say a DICOM read, this object is assigned the
        # result of the read and can be used by a friendly caller.
        self.obj            = None

    def filesFind(self, *args, **kwargs) -> dict:
        """
        This method simply returns a list of files
        and directories down a filesystem tree starting
        from the kwarg:

            root        = <someStartPath>
            fileSubStr  = <someSubStr>

        where the 'fileSubStr' optionally filters all the files
        with <someSubStr>.
        """
        d_ret           : dict  = {
            'status':   False,
            'l_fileFS': [],
            'l_dirFS':  [],
            'numFiles': 0,
            'numDirs':  0
        }
        str_rootPath    : str   = ''
        str_fileSubStr  : str   = ''
        for k,v in kwargs.items():
            if k == 'root'          : str_rootPath      = v
            if k == 'fileSubStr'    : str_fileSubStr    = v
        if len(str_rootPath):
            # Create a list of all files down the <str_rootPath>
            for root, dirs, files in os.walk(str_rootPath):
                for filename in files:
                    d_ret['l_fileFS'].append(os.path.join(root, filename))
                    d_ret['status'] = True
                for dirname in dirs:
                    d_ret['l_dirFS'].append(os.path.join(root, dirname))
            if len(str_fileSubStr):
                d_ret['l_fileFS']   = [
                                        f for f in d_ret['l_fileFS']
                                            if str_fileSubStr in f
                                        ]

        d_ret['numFiles']   = len(d_ret['l_fileFS'])
        d_ret['numDirs']    = len(d_ret['l_dirFS'])
        return d_ret

    @abc.abstractmethod
    def connect(self, *args, **kwargs):
        """
        The base connection class.

        This handles the connection to the openstorage providing service.
        """

    @abc.abstractmethod
    def ls_process(self, *args, **kwargs):
        """
        The base ls process method.

        This handles the ls processing in the openstorage providing service.
        """

    @abc.abstractmethod
    def ls(self, *args, **kwargs):
        """
        Base listing class.

        Provide a listing of resources in the openstorage providing
        service.
        """

    @abc.abstractmethod
    def objExists(self, *args, **kwargs):
        """
        Base object existance class.

        Check if an object exists in the openstorage providing service.
        """

    @abc.abstractmethod
    def objPut(self, *args, **kwargs):
        """
        Base object put method.

        Put a list of (file) objects into storage.
        """

    @abc.abstractmethod
    def objPull(self, *args, **kwargs):
        """
        Base object pull method.

        Pull a list of (file) objects from storage.
        """

class swiftStorage(PfStorage):

    def __init__(self, arg, *args, **kwargs):
        """
        Core initialization and logic in the base class
        """
        # Check if an upstream object exists, and if so
        # merge those args with the current namespace:
        if 'upstream' in arg.keys():
            d_argCopy           = arg.copy()
            # "merge" these 'arg's with upstream.
            arg.update(arg['reportData']['args'])
            # Since this might overwrite some args specific to this
            # app, we update again to the copy.
            arg.update(d_argCopy)

        PfStorage.__init__(self, arg, *args, **kwargs)

    @static_vars(str_prependBucketPath = "")
    def connect(self, *args, **kwargs) -> dict:
        """
        Connect to swift storage and return the connection object,
        as well an optional "prepend" string to fully qualify
        object location in swift storage.

        The 'prependBucketPath' is somewhat 'legacy' to a similar
        method in charm.py and included here with the idea
        to eventually converge on a single swift-based intermediary
        library for both pfcon and CUBE.
        """

        d_ret       = {
            'status':               True,
            'conn':                 None,
            'user':                 self.state('/swift/username'),
            'key':                  self.state('/swift/key'),
            'authurl':              self.state('/swift/auth_url'),
            'container_name':       self.state('/swift/container_name')
        }

        # initiate a swift service connection, based on internal
        # settings already available in the django variable space.
        try:
            d_ret['conn'] = swiftclient.Connection(
                user    = d_ret['user'],
                key     = d_ret['key'],
                authurl = d_ret['authurl']
            )
        except:
            d_ret['status'] = False

        return d_ret

    def rmtree_process(self, *args, **kwargs) -> dict:
        """
        Process the 'rmtree' directive.
        """
        d_ret       = {
            'status':   False,
            'msg':      "No 'meta' JSON directive found in request"
        }
        d_msg       = {}
        d_args      = {}

        for k, v in kwargs.items():
            if k == 'request':      d_msg       = v

        if 'args' in d_msg:
            d_args  = d_msg['args']
            d_ret   = self.rmtree(**d_args)

        return d_ret


    def rmtree(self, **kwargs)  -> dict:
        """
        Remove a "tree" of objects in swift
        """
        d_ret = {
            'status'    : False,
            'dellist'   : []
        }

        d_conn  = self.connect(**kwargs)

        # Get a list of objects
        d_ls    = self.ls(**kwargs)
        if d_ls['status']:
            d_ret['status']     = True
            d_ret['dellist']    = d_ls['list_ls']
            for obj in d_ls['list_ls']:
                d_conn['conn'].delete_object(
                    d_conn['container_name'],
                    obj
                )

        return d_ret

    def ls_process(self, *args, **kwargs) -> dict:
        """
        Process the 'ls' directive (in the appropriate subclass).
        For the case of 'swift', the return dictionary contains a
        key, 'objectDict' containing a list of dictionaries which
        in turn have keys:
            'hash', 'last_modified', 'bytes', 'name', 'content-type'
        """
        d_ret       : dict = {'status': False}
        d_ls        : dict = {}
        d_lsFilter  : dict = {}
        d_msg       : dict = {}
        d_args      : dict = {}
        l_retSpec   : list = ['name', 'bytes', 'hash', 'last_modified']

        for k, v in kwargs.items():
            if k == 'request':      d_msg   = v

        if 'args' in d_msg:
            d_args  = d_msg['args']
            if 'retSpec' in d_args:
                l_retSpec   = d_args['retSpec']
            try:
                d_ls = self.ls(**d_args)
                d_ret['status'] = d_ls['status']
                if len(l_retSpec):
                    d_lsFilter  = [ {x: y[x] for x in l_retSpec}
                                    for y in d_ls['listDict_obj'] ]
                    d_ret['ls'] = d_lsFilter
                else:
                    d_ret['ls'] = d_ls
            except Exception as e:
                d_ret['lsError'] = '%s' % e

        return d_ret

    def ls(self, **kwargs) -> dict:
        """
        Return a dictionary of information about objects in swiftstorage.

        Behaviour specifiers (kwargs):

            path   = <locationInSwift>
            substr      = filter return objects on <substr> in their name
            retSpec     = list of specs to return.
                        Default: ['hash', 'last_modified', 'bytes', 'name']

        Return
        {
            'status':           the status of this call (True/False),
            'listDict_obj':     a list of dictionary objects,
            'list_ls':          a list of object names at the query path,
        }

        """

        l_ls                    : list  = []    # The listing of names to return
        ld_obj                  : list  = []    # List of dictionary objects in swift
        str_path                : str   = '/'
        str_subString           : str   = ''
        b_status                : bool  = False
        l_retSpec               : list  = []

        for k,v in kwargs.items():
            if k == 'path'      : str_path            = v
            if k == 'substr'    : str_subString       = v
            if k == 'retSpec'   : l_retSpec           = v

        # Remove any leading noise on the str_path, specifically
        # any leading '.' characters.
        # This is probably not very robust!
        while str_path[:1] == '.':  str_path    = str_path[1:]

        d_conn          = self.connect(**kwargs)
        if d_conn['status']:
            conn        = d_conn['conn']

            # get the full list of objects in Swift storage with given prefix
            ld_obj = conn.get_container(
                        d_conn['container_name'],
                        prefix          = str_path,
                        full_listing    = True)[1]

            if len(str_subString):
                ld_obj      = [x for x in ld_obj if str_subString in x['name']]

            l_ls            = [x['name'] for x in ld_obj]

            if len(l_retSpec):
                ld_obj      = [ {x: y[x] for x in l_retSpec}
                                    for y in ld_obj ]

            if len(l_ls):   b_status    = True

        return {
            'status':       b_status,
            'listDict_obj': ld_obj,
            'list_ls':      l_ls,
            'conn':         conn
        }

    def objExists(self, **kwargs) -> bool:
        """
        Return True/False if the object at 'path' exists in swift storage.
        """
        b_exists                : bool      = False
        str_obj                 : str       = ''

        for k,v in kwargs.items():
            if k == 'path'      : str_obj   = v

        if len(str_obj):
            kwargs['path']          = str_obj
            d_ls                    = self.ls(**kwargs)
            if d_ls['status']:
                for name in d_ls['list_ls']:
                    if name == str_obj:
                        b_exists = True

        return b_exists

    def objPut_process(self, *args, **kwargs) -> dict:
        """
        Process the 'objPut' directive.

        DICOM handling
        --------------

        A special behaviour is available for DICOM files, triggered by passing
        a kwarg of 'DICOMsubstr = <X>'. In this case, DICOM files (as identi-
        fied by containing the substring pattern within the filename) will be
        read for tag information used to generate the fully qualified storage
        path.

        This fully qualified storage path will be substituted into the
        'toLocation = <someswiftpath>' by replacing the special tag
        '%pack' in the <someswiftpath>.

        NOTE:
        * Typically a list of files all constitute the same DICOM SERIES
          and as such, only one of file in the list needs to be processed for
          packing tags.
        * If the 'do' 'objPut' directive contains a true value for the field
          'packEachDICOM', then each DICOM will be explicitly examined and
          packed individually.

        """

        def toLocation_updateWithDICOMtags(str_DICOMfilename) -> dict:
            """
            Read the str_DICOMfilename, determine the pack path,
            and update the 'toLocation' if necessary.

            Return the original and modified 'toLocation' and status flag.
            """
            b_pack                      = False
            d_DICOMread                 = self.packer.DICOMfile_read(file = str_DICOMfilename)
            d_path                      = self.packer.packPath_resolve(d_DICOMread)
            self.obj[str_DICOMfilename] = d_DICOMread
            str_origTo                  = d_args['toLocation']
            if '%pack' in d_args['toLocation']:
                b_pack  = True
                d_args['toLocation'] = \
                    d_args['toLocation'].replace('%pack', d_path['packDir'])
            return {
                'pack'              : b_pack,
                'originalLocation'  : str_origTo,
                'path'              : d_path,
                'toLocation'        : d_args['toLocation']
            }

        def files_putSingly() -> dict:
            """
            Handle a single file put, return and update d_ret
            """
            nonlocal d_ret
            nonlocal b_singleShot
            d_pack  : dict          = {}
            b_singleShot            = True
            d_ret                   = {
                'status'            : False,
                'localFileList'     : [],
                'objectFileList'    : []
            }
            self.obj                = {}
            # pudb.set_trace()
            for f in d_fileList['l_fileFS']:
                d_pack                  = toLocation_updateWithDICOMtags(f)
                d_args['file']          = f
                d_args['remoteFile']    = d_pack['path']['imageFile']
                d_put                   = self.objPut(**d_args)
                d_ret[f]                = d_put
                d_args['toLocation']    = d_pack['originalLocation']
                d_ret['status']         = d_put['status']
                if d_ret['status']:
                    d_ret['localFileList'].append(d_put['localFileList'][0])
                    d_ret['objectFileList'].append(d_put['objectFileList'][0])
                else:
                    break
            return d_ret

        d_ret           :   dict  = {
            'status'    :   False,
            'msg'       :   "No 'arg' JSON directive found in request"
        }

        d_msg           : dict  = {}
        d_args          : dict  = {}
        str_localPath   : str   = ""
        str_DICOMsubstr : str   = ""
        b_singleShot    : bool  = False

        for k, v in kwargs.items():
            if k == 'request':      d_msg       = v
        # pudb.set_trace()
        if 'args' in d_msg:
            d_args  = d_msg['args']
            if 'localpath' in d_args:
                str_localPath       = d_args['localpath']
                if 'DICOMsubstr' in d_args:
                    d_fileList      = self.filesFind(
                                        root        = str_localPath,
                                        fileSubStr  = d_args['DICOMsubstr']
                                    )
                    if 'packEachDICOM' in d_args:
                        if d_args['packEachDICOM']: files_putSingly()
                    if len(d_fileList['l_fileFS']) and not b_singleShot:
                        toLocation_updateWithDICOMtags(d_fileList['l_fileFS'][0])
                else:
                    d_fileList      = self.filesFind(
                                        root        = str_localPath
                    )
                if d_fileList['status'] and not b_singleShot:
                    d_args['fileList']  = d_fileList['l_fileFS']
                    d_ret               = self.objPut(**d_args)
                elif not d_fileList['status']:
                    d_ret['msg']    = 'No valid file list generated'
        return d_ret

    def objPut(self, *args, **kwargs) -> dict:
        """
        Put an object (or list of objects) into swift storage.

        This method also "maps" tree locations in the local storage
        to new locations in the object storage. For example, assume
        a list of local locations starting with:

                /home/user/project/data/ ...

        and we want to pack everything in the 'data' dir to
        object storage, at location '/storage'. In this case, the
        pattern of kwargs specifying this would be:

                fileList = ['/home/user/project/data/file1',
                            '/home/user/project/data/dir1/file_d1',
                            '/home/user/project/data/dir2/file_d2'],
                toLocation      = '/storage',
                mapLocationOver = '/home/user/project/data'

        will replace, for each file in <fileList>, the <mapLocationOver> with
        <inLocation>, resulting in a new list

                '/storage/file1',
                '/storage/dir1/file_d1',
                '/storage/dir2/file_d2'

        """
        b_status                : bool  = True
        l_localfile             : list  = []    # Name on the local file system
        l_remotefileName        : list  = []    # A replacement for the remote filename
        l_objectfile            : list  = []    # Name in the object storage
        str_swiftLocation       : str   = ''
        str_mapLocationOver     : str   = ''
        str_localfilename       : str   = ''
        str_storagefilename     : str   = ''
        str_swiftLocation       : str   = ""
        str_remoteFile          : str   = ""
        d_ret                   : dict  = {
                                            'status':           b_status,
                                            'localFileList':    [],
                                            'objectFileList':   [],
                                            'localpath':        ''
                                        }

        d_conn  = self.connect(*args, **kwargs)

        for k,v in kwargs.items():
            if k == 'file'              : l_localfile.append(v)
            if k == 'remoteFile'        : l_remotefileName.append(v)
            if k == 'remoteFileList'    : l_remotefileName      = v
            if k == 'fileList'          : l_localfile           = v
            if k == 'toLocation'        : str_swiftLocation     = v
            if k == 'mapLocationOver'   : str_mapLocationOver   = v

        if len(str_mapLocationOver):
            # replace the local file path with object store path
            l_objectfile    = [w.replace(str_mapLocationOver, str_swiftLocation) \
                                for w in l_localfile]
        else:
            # Prepend the swiftlocation to each element in the localfile list:
            l_objectfile    = [str_swiftLocation + '{0}'.format(i) for i in l_localfile]

        # Check and possibly change the actual file *names* to put into swift storage
        # (the default is to use the same name as the local file -- however in the
        # case of DICOM files, the actual final file name might also change)
        if len(l_remotefileName):
            l_objectfile    = [l.replace(os.path.basename(l), f) for l,f in
                                    zip(l_objectfile, l_remotefileName)]

        d_ret['localpath']  = os.path.dirname(l_localfile[0])

        if d_conn['status']:
            for str_localfilename, str_storagefilename in zip(l_localfile, l_objectfile):
                try:
                    d_ret['status'] = True and d_ret['status']
                    with open(str_localfilename, 'rb') as fp:
                        d_conn['conn'].put_object(
                            d_conn['container_name'],
                            str_storagefilename,
                            contents=fp.read()
                        )
                except Exception as e:
                    d_ret['error']  = '%s' % e
                    d_ret['status'] = False
                d_ret['localFileList'].append(str_localfilename)
                d_ret['objectFileList'].append(str_storagefilename)
        return d_ret

    def objPull_process(self, *args, **kwargs):
        """
        Process the 'objPull' directive.
        """
        d_ret       :   dict  = {
            'status':   False,
            'msg'   :   "No 'meta' JSON directive found in request"
        }
        d_msg       = {}
        d_args      = {}

        for k, v in kwargs.items():
            if k == 'request':      d_msg       = v

        if 'args' in d_msg:
            d_args  = d_msg['args']
            d_ret   = self.objPull(**d_args)

        return d_ret

    def objPull(self, *args, **kwargs):
        """
        Pull an object (or set of objects) from swift storage and
        onto the local filesystem.

        This method can also "map" locations in the object storage
        to new locations in the filesystem storage. For example, assume
        a list of object locations starting with:

                user/someuser/uploads/project/data ...

        and we want to pack everything from 'data' to the local filesystem
        to, for example,

                /some/dir/data

        In this case, the pattern of kwargs specifying this would be:

                    fromLocation    = user/someuser/uploads/project/data
                    toLocation      = /some/dir/data

        if 'toLocation' is not specified, then the local file system
        location will be the 'fromLocation' prefixed with a '/'.

        """
        b_status                : bool  = True
        l_localfile             : list  = []    # Name on the local file system
        l_objectfile            : list  = []    # Name in the object storage
        str_swiftLocation       : str   = ''
        str_mapLocationOver     : str   = ''
        str_localfilename       : str   = ''
        str_storagefilename     : str   = ''
        str_swiftLocation       : str   = ""
        d_ret                   = {
            'status':           b_status,
            'localFileList':    [],
            'objectFileList':   [],
            'localpath':        ''
        }

        d_conn  = self.connect(*args, **kwargs)

        for k,v in kwargs.items():
            if k == 'fromLocation'  : str_swiftLocation   = v
            if k == 'toLocation'    : str_mapLocationOver = v

        kwargs['path']  = str_swiftLocation
        # Get dictionary of objects in storage
        d_ls            = self.ls(*args, **kwargs)

        # List of objects in storage
        l_objectfile    = [x['name'] for x in d_ls['listDict_obj']]

        if len(str_mapLocationOver):
            # replace the local file path with object store path
            l_localfile         = [w.replace(str_swiftLocation, str_mapLocationOver) \
                                    for w in l_objectfile]
        else:
            # Prepend a '/' to each element in the l_objectfile:
            l_localfile         = ['/' + '{0}'.format(i) for i in l_objectfile]
            str_mapLocationOver =  '/' + str_swiftLocation

        d_ret['localpath']          = str_mapLocationOver
        d_ret['currentWorkingDir']  = os.getcwd()

        if d_conn['status']:
            for str_localfilename, str_storagefilename in zip(l_localfile, l_objectfile):
                try:
                    d_ret['status'] = True and d_ret['status']
                    obj_tuple       = d_conn['conn'].get_object(
                                                    d_conn['container_name'],
                                                    str_storagefilename
                                                )
                    str_parentDir   = os.path.dirname(str_localfilename)
                    os.makedirs(str_parentDir, exist_ok = True)
                    with open(str_localfilename, 'wb') as fp:
                        # fp.write(str(obj_tuple[1], 'utf-8'))
                        fp.write(obj_tuple[1])
                except Exception as e:
                    d_ret['error']  = str(e)
                    d_ret['status'] = False
                d_ret['localFileList'].append(str_localfilename)
                d_ret['objectFileList'].append(str_storagefilename)
        return d_ret

    def run(self, opt={}) -> dict:
        """
        Perform the storage operation
        """
        d_actionResult  : dict  = {
            'status'    : False,
            'msg'       : ''
        }
        try:
            # First see if the "do" directive is a CLI
            # flag captured in the self.arg structure
            d_do            : dict  = json.loads(self.arg['do'])
        except:
            # Else, assume that the d_do is the passed opt
            d_do            = opt
        if 'action' in d_do:
            self.log("verb: %s detected." % d_do['action'],
                      comms = 'status')
            str_method      = '%s_process' % d_do['action']
            self.log("method to call: %s(request = d_msg) " % str_method,
                      comms = 'status')
            try:
                # pudb.set_trace()
                method              = getattr(self, str_method)
                d_actionResult      = method(request = d_do)
            except:
                str_msg     = "Class '{}' does not implement method '{}'".format(
                                        self.__class__.__name__,
                                        str_method)
                d_actionResult      = {
                    'status':   False,
                    'msg':      str_msg
                }
                self.log(str_msg, comms = 'error')
            self.log(json.dumps(d_actionResult, indent = 4), comms = 'tx')

        return d_actionResult


# Global modules
import  subprocess, re, collections
import  asyncio
from typing import Dict, TypedDict

import  pudb
import  json
import  os
import  sys
from    datetime            import  datetime
from    dateutil            import  relativedelta
from    terminaltables      import  SingleTable
from    argparse            import  Namespace, ArgumentParser
from    argparse            import  RawTextHelpFormatter
import  time

import  pfmisc
from    pfmisc._colors      import  Colors

from    dask                import  delayed, compute

# PYPX modules
from    .base               import Base
from    .move               import Move
import  pypx
import  pypx.re
from    pypx                import smdb
from    pypx                import report
from    pypx                import do
import  copy


def parser_setup(str_desc):
    parser = ArgumentParser(
                description         = str_desc,
                formatter_class     = RawTextHelpFormatter
            )

    # JSONarg
    parser.add_argument(
        '--JSONargs',
        action  = 'store',
        dest    = 'JSONargString',
        type    = str,
        default = '',
        help    = 'JSON equivalent of CLI key/values')


    # db access settings
    parser.add_argument(
        '--db',
        action  = 'store',
        dest    = 'dblogbasepath',
        type    = str,
        default = '/tmp/log',
        help    = 'path to base dir of receipt database')

    # service access settings
    parser.add_argument(
        '--PACS',
        action  = 'store',
        dest    = 'PACS',
        type    = str,
        default = '',
        help    = 'PACS lookup service identifier')
    parser.add_argument(
        '--CUBE',
        action  = 'store',
        dest    = 'CUBE',
        type    = str,
        default = '',
        help    = 'CUBE lookup service identifier')
    parser.add_argument(
        '--swift',
        action  = 'store',
        dest    = 'swift',
        type    = str,
        default = '',
        help    = 'swift lookup service identifier')


    # PACS access settings
    parser.add_argument(
        '--aet',
        action  = 'store',
        dest    = 'aet',
        type    = str,
        default = 'CHRIS-ULTRON-AET',
        help    = 'aet')
    parser.add_argument(
        '--aec',
        action  = 'store',
        dest    = 'aec',
        type    = str,
        default = 'CHRIS-ULTRON-AEC',
        help    = 'aec')
    parser.add_argument(
        '--serverIP',
        action  = 'store',
        dest    = 'serverIP',
        type    = str,
        default = '192.168.1.110',
        help    = 'PACS server IP')
    parser.add_argument(
        '--serverPort',
        action  = 'store',
        dest    = 'serverPort',
        type    = str,
        default = '4242',
        help    = 'PACS server port')
    parser.add_argument(
        '--findscu',
        action  = 'store',
        dest    = 'findscu',
        type    = str,
        default = '/usr/bin/findscu',
        help    = '"findscu" executable absolute location')
    parser.add_argument(
        '--movescu',
        action  = 'store',
        dest    = 'movescu',
        type    = str,
        default = '/usr/bin/movescu',
        help    = '"movescu" executable absolute location')

    # Query settings
    parser.add_argument(
        '--AccessionNumber',
        action  = 'store',
        dest    = 'AccessionNumber',
        type    = str,
        default = '',
        help    = 'Accession Number')
    parser.add_argument(
        '--PatientID',
        action  = 'store',
        dest    = 'PatientID',
        type    = str,
        default = '',
        help    = 'Patient ID')
    parser.add_argument(
        '--PatientName',
        action  = 'store',
        dest    = 'PatientName',
        type    = str,
        default = '',
        help    = 'Patient name')
    parser.add_argument(
        '--PatientSex',
        action  = 'store',
        dest    = 'PatientSex',
        type    = str,
        default = '',
        help    ='Patient sex')
    parser.add_argument(
        '--StudyDate',
        action  = 'store',
        dest    = 'StudyDate',
        type    = str,
        default = '',
        help    = 'Study date (YYYY/MM/DD)')
    parser.add_argument(
        '--ModalitiesInStudy',
        action  = 'store',
        dest    = 'ModalitiesInStudy',
        type    = str,
        default = '',
        help    = 'Modalities in study')
    parser.add_argument(
        '--Modality',
        action  = 'store',
        dest    = 'Modality',
        type    = str,
        default = '',
        help    = 'Study Modality')
    parser.add_argument(
        '--PerformedStationAETitle',
        action  = 'store',
        dest    = 'PerformedStationAETitle',
        type    = str,
        default = '',
        help    = 'Performed station aet')
    parser.add_argument(
        '--StudyDescription',
        action  = 'store',
        dest    = 'StudyDescription',
        type    = str,
        default = '',
        help    = 'Study description')
    parser.add_argument(
        '--SeriesDescription',
        action  = 'store',
        dest    = 'SeriesDescription',
        type    = str,
        default = '',
        help    = 'Series Description')
    parser.add_argument(
        '--SeriesInstanceUID',
        action  = 'store',
        dest    = 'SeriesInstanceUID',
        type    = str,
        default = '',
        help    = 'Series Instance UID')
    parser.add_argument(
        '--StudyInstanceUID',
        action  = 'store',
        dest    = 'StudyInstanceUID',
        type    = str,
        default = '',
        help    = 'Study Instance UID')
    parser.add_argument(
        '--ProtocolName',
        action  = 'store',
        dest    = 'ProtocolName',
        type    = str,
        default = '',
        help    = 'Protocol Name')
    parser.add_argument(
        '--AcquisitionProtocolName',
        action  = 'store',
        dest    = 'AcquisitionProtocolName',
        type    = str,
        default = '',
        help    = 'Acquisition Protocol Description Name')

    parser.add_argument(
        '--AcquisitionProtocolDescription',
        action  = 'store',
        dest    = 'AcquisitionProtocolDescription',
        type    = str,
        default = '',
        help    = 'Acquisition Protocol Description')

    # Retrieve settings
    parser.add_argument(
        '--withFeedBack',
        action  = 'store_true',
        dest    = 'withFeedBack',
        default = False,
        help    = 'If specified, print the "then" events as they happen')
    parser.add_argument(
        '--then',
        action  = 'store',
        dest    = 'then',
        default = "",
        help    = 'If specified, perform another set operations "next" after the find')
    parser.add_argument(
        '--thenArgs',
        action  = 'store',
        dest    = 'thenArgs',
        default = "",
        help    = 'If specified, associate the corresponding JSON string in the list to a then operation')
    parser.add_argument(
        '--intraSeriesRetrieveDelay',
        action  = 'store',
        dest    = 'intraSeriesRetrieveDelay',
        default = "0",
        help    = 'If specified, then wait specified seconds between retrieve series loops')
    parser.add_argument(
        '--move',
        action  = 'store_true',
        dest    = 'move',
        default = False,
        help    = 'If specified with --retrieve, call initiate a PACS pull on the set of SeriesUIDs using pypx/move')

    parser.add_argument(
        '--json',
        action  = 'store_true',
        dest    = 'json',
        default = False,
        help    = 'If specified, dump the JSON structure relating to the query')

    parser.add_argument(
        "-v", "--verbosity",
        help    = "verbosity level for app",
        dest    = 'verbosity',
        type    = int,
        default = 1)
    parser.add_argument(
        "-x", "--desc",
        help    = "long synopsis",
        dest    = 'desc',
        action  = 'store_true',
        default = False
    )
    parser.add_argument(
        "-y", "--synopsis",
        help    = "short synopsis",
        dest    = 'synopsis',
        action  = 'store_true',
        default = False
    )
    parser.add_argument(
        '--version',
        help    = 'if specified, print version number',
        dest    = 'b_version',
        action  = 'store_true',
        default = False
    )
    parser.add_argument(
        '--waitForUserTerminate',
        help    = 'if specified, wait for user termination',
        dest    = 'b_waitForUserTerminate',
        action  = 'store_true',
        default = False
    )
    parser.add_argument(
        '--reallyEfficient',
        help    = 'record NumberOfSeriesRelatedInstances to redis',
        dest    = 'b_reallyEfficient',
        action  = 'store_true',
        default = False
    )

    return parser

def parser_interpret(parser, *args):
    """
    Interpret the list space of *args, or sys.argv[1:] if
    *args is empty
    """
    if len(args):
        args    = parser.parse_args(*args)
    else:
        args    = parser.parse_args(sys.argv[1:])
    return args

def parser_JSONinterpret(parser, d_JSONargs):
    """
    Interpret a JSON dictionary in lieu of CLI.

    For each <key>:<value> in the d_JSONargs, append to
    list two strings ["--<key>", "<value>"] and then
    argparse.
    """
    l_args  = []
    for k, v in d_JSONargs.items():
        if type(v) == type(True):
            if v: l_args.append('--%s' % k)
            continue
        l_args.append('--%s' % k)
        l_args.append('%s' % v)
    return parser_interpret(parser, l_args)

class Find(Base):

    """
    The Find module provides rather extensive PACS query
    functionality.

    See the 'query' method for the space of query parameters
    that the module offers. Text data pertaining to image sets
    that match the pattern of query values are returned by
    this method to a called as a JSON/dictionary payload.

    The return dictionary contains a field, 'report' that itself
    contains three report formats: 'tabular', 'rawText', and 'json'.
    The 'tabular' and 'rawText' reports are for console
    consumption/presentation, while the 'json' report is for
    software agents.

    This 'json' report also returns a hidden 'bodySeriesUID'
    section, with each entry corresponding in order to the
    seriesDescription that the report returns. These values
    are the SeriesInstanceUIDs that can actually be retrieved
    by the pypx/move operation.
    """

    def __init__(self, arg: Dict):
        """
        Constructor.

        Defines a default report structure, divided into a
        "header" and a "body". In each section,  DICOM tags
        retrieved from either the STUDY or SERIES level are
        catalogued.

        Since a given STUDY typically has several SERIES, in
        most cases only SERIES level tags are included in the
        "body".

        In some cases, some tags are only available at
        the SERIES level (such as the Modality). If such a tag
        is in the STUDY level, then the corresponding tag from
        the FIRST series in the STUDY is reported.
        """
        super(Find, self).__init__(arg)
        self.dp             = pfmisc.debug(
                                        verbosity   = self.verbosity,
                                        within      = 'Find',
                                        syslog      = False
                                        )
        self.log            = self.dp.qprint
        self.then           = do.Do(self.arg)
        self.reallyEfficient = arg.get('b_reallyEfficient', False)

    def query(self, opt={}):
        parameters = {
            'AccessionNumber':                  '',
            'PatientID':                        '',
            'PatientName':                      '',
            'PatientBirthDate':                 '',
            'PatientAge':                       '',
            'PatientSex':                       '',
            'StudyDate':                        '',
            'StudyDescription':                 '',
            'StudyInstanceUID':                 '',
            'Modality':                         '',
            'ModalitiesInStudy':                '',
            'PerformedStationAETitle':          '',
            'NumberOfPatientRelatedInstances':  '',
            'NumberOfPatientRelatedStudies':    '',
            'NumberOfPatientRelatedSeries':     '',
            'NumberOfStudyRelatedInstances':    '',
            'NumberOfStudyRelatedSeries':       '',
            'NumberOfSeriesRelatedInstances':   '',
            'InstanceNumber':                   '',
            'SeriesDate':                       '',
            'SeriesDescription':                '',
            'SeriesInstanceUID':                '',
            'ProtocolName':                     '',
            'AcquisitionProtocolDescription':   '',
            'AcquisitionProtocolName':          '',
            'QueryRetrieveLevel':               'SERIES'
        }

        query = ''
        # we use a sorted dictionary so we can test generated command
        # more easily
        ordered = collections.OrderedDict(
                        sorted(
                                parameters.items(),
                                key=lambda t: t[0]
                                )
                        )
        for key, value in ordered.items():
            # update value if provided
            if key in opt:
                value = opt[key]
            # update query
            if value != '':
                query += ' -k "' + key + '=' + value + '"'
            else:
                query += ' -k ' + key

        return query

    def xinetd_command(self, opt={}):
        return "service xinetd restart"

    def findscu_command(self, opt={} ):

        command = '-xi -S'
        str_cmd     = "%s %s %s %s" % (
                        self.findscu,
                        command,
                        self.query(opt),
                        self.commandSuffix()
        )
        return str_cmd

    async def run(self, opt={}):
        """
        Main entry method.

        In order to accommodate the widest range of PACS dialects,
        a query occurs in two phases/passes:

            * First, at the STUDY level to receive the set of possible
              StudyUIDs
            * Then, given each STUDY, run at the SERIES level to
              receive the set of SeriesUID information

        The query itself is based on the pattern of DICOM tag specifications
        given used to instantiate this class.

        """
        # First we execute on a STUDY level to determine all the
        # STUDIES related to this query
        series_uid                  = opt["SeriesInstanceUID"]
        opt["SeriesInstanceUID"]    = ""
        formattedStudiesResponse    = \
            await self.systemlevel_runasync(opt,
                    {
                        'f_commandGen':         self.findscu_command,
                        'QueryRetrieveLevel':   'STUDY'
                    }
            )

        if formattedStudiesResponse['status']  != 'error':
            filteredStudiesResponse             = {}
            filteredStudiesResponse['status']   = formattedStudiesResponse['status']
            filteredStudiesResponse['command']  = formattedStudiesResponse['command']
            filteredStudiesResponse['data']     = []
            filteredStudiesResponse['args']     = self.arg
            studyIndex                          = 0
            for study in formattedStudiesResponse['data']:
                l_seriesResults = []
                # For each study, we now execute a query on a SERIES
                # level to complete the picture.
                formattedSeriesResponse     = \
                    await self.systemlevel_runasync(opt,
                            {
                                'f_commandGen':         self.findscu_command,
                                'QueryRetrieveLevel':   'SERIES',
                                'StudyInstanceUID':     study['StudyInstanceUID']['value'],
                                'SeriesInstanceUID':    series_uid
                            }
                    )
                for series in formattedSeriesResponse['data']:
                    series['label']             = {}
                    series['label']['tag']      = 0
                    series['label']['value']    = "SERIES"
                    series['label']['label']    = 'RetrieveLevel'

                    series['command']           = {}
                    series['command']['tag']    = 0
                    series['command']['value']  = formattedSeriesResponse['command']
                    series['command']['label']  = 'command'

                    series['status']            = {}
                    series['status']['tag']     = 0
                    series['status']['value']   = formattedSeriesResponse['status']
                    series['status']['label']   = 'status'

                    l_seriesResults.append(series)

                if len(l_seriesResults):
                    filteredStudiesResponse['data'].append(study)
                    filteredStudiesResponse['data'][-1]['series']           \
                        = l_seriesResults

                formattedStudiesResponse['data'][studyIndex]['series']      \
                    = l_seriesResults
                studyIndex+=1

            if self.reallyEfficient:
                await self.recordInRedis(filteredStudiesResponse['data'])

            if len(self.arg['then']):
                self.then.arg['reportData']     = copy.deepcopy(filteredStudiesResponse)
                d_then                          = await self.then.run()
                filteredStudiesResponse['then'] = copy.deepcopy(d_then)

            return filteredStudiesResponse
        else:
            return formattedStudiesResponse

    async def recordInRedis(self, data):
        """
        Record the value of ``NumberOfSeriesRelatedInstances`` for each series to redis.
        """
        seriesInfo = self._extractSeriesRelatedInstances(data)
        connection = await pypx.re.getRedisClient()
        async with connection.pipeline(transaction=True) as pipe:
            for k, v in seriesInfo.items():
                if (await connection.hget(k, 'fileCounter')) is None:
                    # TODO we also want to avoid writing to redis when series is already packed
                    pipe.hset(k, mapping=v)
            await pipe.execute()
        await connection.close()

    @classmethod
    def _extractSeriesRelatedInstances(cls, data) -> Dict[str, pypx.re.ReData]:
        timestamp = datetime.now().isoformat()
        all_series = (s for a in data for s in a['series'])
        return {
            cls._redisKeyOf(series): cls._createReData(series, timestamp)
            for series in all_series
            if cls._seriesHasNumberofSeriesRelatedInstances(series)
        }

    @staticmethod
    def _redisKeyOf(s: dict) -> str:
        return pypx.re.redisKeyOf(s['SeriesInstanceUID']['value'])

    @staticmethod
    def _seriesHasNumberofSeriesRelatedInstances(s: dict) -> bool:
        try:
            num = int(s['NumberOfSeriesRelatedInstances']['value'])
            return num >= 1
        except (KeyError, ValueError):
            return False

    @staticmethod
    def _createReData(s: dict, date: str) -> pypx.re.ReData:
        """``ReData`` constructor"""
        return {
            'NumberOfSeriesRelatedInstances': s['NumberOfSeriesRelatedInstances']['value'],
            'fileCounter': 0,
            'lastUpdate': date
        }



# Global modules
# import  argparse
# import  subprocess, re, collections
from    pfmisc.other import list_removeDuplicates
import  pudb
import  json

import  asyncio

from    datetime            import  datetime
from    argparse            import  Namespace, ArgumentParser
from    argparse            import  RawTextHelpFormatter
import  time

import  pfmisc
from    pfmisc._colors      import  Colors

from    dask                import  delayed, compute
import  sys

# PYPX modules
from    .base               import Base
from    .move               import Move
import  pypx
from    pypx                import smdb
from    pypx                import report
from    pypx.push           import parser_setup         as pushParser_setup
from    pypx.push           import parser_JSONinterpret as pushParser_JSONinterpret
from    pypx.register       import parser_setup         as registerParser_setup
from    pypx.register       import parser_JSONinterpret as registerParser_JSONinterpret


def parser_setup(str_desc):
    parser = ArgumentParser(
                description         = str_desc,
                formatter_class     = RawTextHelpFormatter
            )

    # the main report data to interpret
    parser.add_argument(
        '--reportData',
        action  = 'store',
        dest    = 'reportData',
        type    = str,
        default = '',
        help    = 'JSON report from upstream process')

    parser.add_argument(
        '--reportDataFile',
        action  = 'store',
        dest    = 'reportDataFile',
        type    = str,
        default = '',
        help    = 'JSON report contained in file from upstream process')

    # db access settings
    parser.add_argument(
        '--db',
        action  = 'store',
        dest    = 'dblogbasepath',
        type    = str,
        default = '/tmp/log',
        help    = 'path to base dir of receipt database')

    # Behaviour settings
    parser.add_argument(
        '--withFeedBack',
        action  = 'store_true',
        dest    = 'withFeedBack',
        default = False,
        help    = 'If specified, print the "then" events as they happen')
    parser.add_argument(
        '--then',
        action  = 'store',
        dest    = 'then',
        default = "",
        help    = 'If specified, then perform the set of operations')
    parser.add_argument(
        '--thenArgs',
        action  = 'store',
        dest    = 'thenArgs',
        default = "",
        help    = 'If specified, associate the corresponding JSON string in the list to a then operation')
    parser.add_argument(
        '--intraSeriesRetrieveDelay',
        action  = 'store',
        dest    = 'intraSeriesRetrieveDelay',
        default = "0",
        help    = 'If specified, then wait specified seconds between retrieve series loops')

    parser.add_argument(
        '--move',
        action  = 'store_true',
        dest    = 'move',
        default = False,
        help    = 'If specified with --retrieve, call initiate a PACS pull on the set of SeriesUIDs using pypx/move')

    parser.add_argument(
        '--json',
        action  = 'store_true',
        dest    = 'json',
        default = False,
        help    = 'If specified, dump the JSON structure relating to the query')

    parser.add_argument(
        "-v", "--verbosity",
        help    = "verbosity level for app",
        dest    = 'verbosity',
        type    = int,
        default = 1)
    parser.add_argument(
        "-x", "--desc",
        help    = "long synopsis",
        dest    = 'desc',
        action  = 'store_true',
        default = False
    )
    parser.add_argument(
        "-y", "--synopsis",
        help    = "short synopsis",
        dest    = 'synopsis',
        action  = 'store_true',
        default = False
    )
    parser.add_argument(
        '--version',
        help    = 'if specified, print version number',
        dest    = 'b_version',
        action  = 'store_true',
        default = False
    )
    parser.add_argument(
        '--waitForUserTerminate',
        help    = 'if specified, wait for user termination',
        dest    = 'b_waitForUserTerminate',
        action  = 'store_true',
        default = False
    )

    return parser

def parser_interpret(parser, *args):
    """
    Interpret the list space of *args, or sys.argv[1:] if
    *args is empty
    """
    if len(args):
        args    = parser.parse_args(*args)
    else:
        args    = parser.parse_args(sys.argv[1:])
    return args

def parser_JSONinterpret(parser, d_JSONargs):
    """
    Interpret a JSON dictionary in lieu of CLI.

    For each <key>:<value> in the d_JSONargs, append to
    list two strings ["--<key>", "<value>"] and then
    argparse.
    """
    l_args  = []
    for k, v in d_JSONargs.items():
        l_args.append('--%s' % k)
        if type(v) == type(True): continue
        l_args.append('%s' % v)
    return parser_interpret(parser, l_args)

class Do(Base):

    """
    The Do module provides a convient and centralised location
    from which to dispatch processing to a variety of additional
    operations.

    All these operations work on a study/series tuple, and this
    module iterates over structures that define this space.

    In many cases, the operations append to the JSON stream that
    "flows" through this module, adding some operation-specific
    information, which in turn can be consumed by a downstream
    process.

    """

    def __init__(self, arg):
        """
        Constructor. Nothing too fancy. Just a parent init
        and a class logging function.

        Most of the "complexity" here is "merging" upstream
        arg values with this specific app's arg values.

        """
        # Check if an upstream 'reportData' exists, and if so
        # merge those the upstream process's CLI args into the
        # current namespace.
        #
        # NOTE:
        #
        # * the merge is on the *dest* of the argparse namespace, not
        #   the CLI keys -- so on 'b_json' and not '--json' for example.
        #
        # * this merge WILL OVERWRITE/CLOBBER any CLI specified
        #   for this app in favor of upstream ones *except* for
        #   the 'withFeedBack' and 'json'!
        #
        # * CLI dest keys that are not in the CLI space of this app
        #   are nonetheless still added to the arg structure -- this
        #   allows for downstream tranmission to apps with different
        #   CLI dest spaces.

        # pudb.set_trace()
        if 'reportData' in arg.keys():
            if 'args' in arg['reportData']:
                for k,v in arg['reportData']['args'].items():
                    # if k in arg and len('%s' % v):
                    if len('%s' % v):
                        if k not in ['json', 'withFeedBack']:
                            arg[k] = v

        # Minor "sanity" check... if 'withFeedBack' is True
        # then 'json' should be False
        if arg['withFeedBack']: arg['json'] = False

        super(Do, self).__init__(arg)
        self.dp             = pfmisc.debug(
                                        verbosity   = self.verbosity,
                                        within      = 'Do',
                                        syslog      = False
                                        )
        self.log            = self.dp.qprint

    async def run(self, opt={}) -> dict:
        """
        If a '--next' request has been specified in the class / module,
        perform additional actions on the space of find hits.

        These actions are typically either:

            * retrieve images
            * status query on retrieved images
            * find query on the internal smdb "database"
            * push images to a swift/CUBE
            * register images from a push into CUBE

        and are performed with additional processing on the STUDY/SERIES
        level. This essentially loops over all the SeriesInstanceUID in the
        query space structure.

        NOTE: The architecture/infrastructure could be overwhelmed if
        too many concurrent requests are presented. Note that a separate
        `storescp` is spawned for EACH incoming DICOM file. Thus requesting
        multiple series (each with multiple DICOM files) in multiple studies
        all at once could result in thousands of `storescp` being spawned
        to try and handle the flood.
        """

        def countDownTimer_do(f_time):
            t               : int   = int(f_time)
            while t:
                mins, secs = divmod(t, 60)
                timer = '{:02d}:{:02d}'.format(mins, secs)
                if self.arg['withFeedBack']:
                    print("     ", end = '')
                    print(  Colors.BLUE_BCKGRND + Colors.WHITE +          \
                            "[ Parsing incoming images %s ]" % timer + Colors.NO_COLOUR
                    )
                time.sleep(1)
                if self.arg['withFeedBack']: print("\033[2A")
                t -= 1

        def seriesRetrieveDelay_do(str_line):
            """
            Simply delay processing a retrieve to prevent client
            overwhelm.

            Delay can be a simple fixed interval in (float) seconds,
            or if the delay is the string "dynamic" then delay by
            a function of the number of images retrieved.
            """
            factor  = 1
            f_sleep = 0.0
            if 'intraSeriesRetrieveDelay' in self.arg.keys():
                if 'dynamic' not in self.arg['intraSeriesRetrieveDelay']:
                    f_sleep = float(self.arg['intraSeriesRetrieveDelay'])
                else:
                    l_dynamic   = self.arg['intraSeriesRetrieveDelay'].split(':')
                    if len(l_dynamic) ==2:
                        factor  = int(l_dynamic[1])
                    l_words     = str_line.split()
                    images  = int(l_words[1])
                    f_sleep = float(images) / factor
                countDownTimer_do(f_sleep)

        def retrieve_do() -> dict:
            """
            Nested retrieve handler
            """
            nonlocal series, studyIndex, seriesIndex
            seriesInstances : int   = series['NumberOfSeriesRelatedInstances']['value']
            d_then          : dict  = {}
            d_db            : dict  = {}
            d_db                    = db.seriesData(
                                            'retrieve',
                                            'NumberOfSeriesRelatedInstances',
                                            seriesInstances
                                        )
            if d_db['status']:
                str_line        = presenter.seriesRetrieve_print(
                    studyIndex  = studyIndex, seriesIndex = seriesIndex
                )
                if self.arg['withFeedBack']: self.log(str_line + "               ")
                series['SeriesMetaDescription']  = {
                                        'tag'   : "0,0",
                                        'value' : str_line,
                                        'label' : 'inlineRetrieveText'
                                    }
                if self.move:
                    d_then      = pypx.move({
                                    **self.arg,
                                })
                else:
                    d_then = self.systemlevel_run(self.arg,
                            {
                                'f_commandGen'      : self.movescu_command,
                                'series_uid'        : str_seriesUID,
                                'study_uid'         : str_studyUID
                            }
                    )
                if 'intraSeriesRetrieveDelay' in self.arg.keys():
                    if self.arg['intraSeriesRetrieveDelay']:
                        seriesRetrieveDelay_do(str_line)
                series['PACS_retrieve'] = {
                    'requested' :   '%s' % datetime.now()
                }
                d_db    = db.seriesData('retrieve', 'command', d_then)
            else:
                if self.arg['withFeedBack']:
                    self.log(d_db['error'])
                d_then  = d_db
            return d_then

        def status_do() -> dict:
            """
            Nested status handler
            """
            nonlocal    series
            d_then      : dict  = {}

            # pudb.set_trace()
            self.arg['verifySeriesInStudy']   = True
            d_then      = pypx.status({
                            **self.arg,
                        })

            str_line    = presenter.seriesStatus_print(
                studyIndex  = studyIndex,
                seriesIndex = seriesIndex,
                status      = d_then
            )
            series['SeriesMetaDescription']    = {
                                    'tag'   : "0,0",
                                    'value' : str_line,
                                    'label' : 'inlineStatusText'
                                }
            if self.arg['withFeedBack']: self.log(str_line)

            return d_then

        def report_do() -> dict:
            """
            Nested simple "echo/show" handler -- can be used as a 'then' in lieu
            of piping into the `px-report` module.
            """
            nonlocal    series
            d_then              : dict  = {}
            str_instanceNumber  : str    = series['InstanceNumber']['value']
            str_line    = presenter.seriesReport_print(
                studyIndex      = studyIndex,
                seriesIndex     = seriesIndex,
                instanceNumber  = str_instanceNumber
            )
            if self.arg['withFeedBack']: self.log(str_line)

            return d_then

        def push_do(d_pushArgsCLI) -> dict:
            """
            Nested push handler
            """
            nonlocal    series
            str_seriesInstanceUID   : str   = series['SeriesInstanceUID']['value']
            str_line                : str   = presenter.seriesPush_print(
                studyIndex  = studyIndex,
                seriesIndex = seriesIndex,
                fileCounts  = db.series_receivedAndRequested(str_seriesInstanceUID)
            )

            if self.arg['withFeedBack']: self.log(str_line)

            d_then                  : dict  = {}
            d_seriesDir             : dict  = db.imageDirs_getOnSeriesInstanceUID(str_seriesInstanceUID)
            pushParser                      = pushParser_setup("Push Parser")
            d_pushArgs                      = pushParser_JSONinterpret(pushParser, d_pushArgsCLI)
            d_pushArgs.str_xcrdir           = d_seriesDir[str_seriesInstanceUID]
            d_then      = pypx.push(
                            vars(d_pushArgs),
                        )

            return d_then

        def register_do(d_registerArgsCLI) -> dict:
            """
            Nested register handler
            """
            nonlocal    series
            str_seriesInstanceUID   : str   = series['SeriesInstanceUID']['value']
            str_line                : str   = presenter.seriesRegister_print(
                studyIndex  = studyIndex,
                seriesIndex = seriesIndex,
                fileCounts  = db.series_receivedAndRequested(str_seriesInstanceUID)
            )

            if self.arg['withFeedBack']: self.log(str_line)
            d_then                  : dict  = {}
            d_seriesDir             : dict  = db.imageDirs_getOnSeriesInstanceUID(str_seriesInstanceUID)
            registerParser                  = registerParser_setup("Register Parser")
            d_registerArgs                  = pushParser_JSONinterpret(registerParser, d_registerArgsCLI)
            d_registerArgs.str_xcrdir       = d_seriesDir[str_seriesInstanceUID]
            d_then      = pypx.register(
                            d_registerArgs,
                        )
            try:
                str_msg     = next(gen_dict_extract('msg', d_then))
            except:
                str_msg     = ""
            if len(str_msg):
                try:
                    self.log(json.dumps(json.loads(str_msg), indent = 4), comms = 'error')
                except:
                    self.log(json.dumps(d_seriesDir, indent = 4), comms = 'error')
                    self.log(str_msg, comms = 'error')
            return d_then

        def gen_dict_extract(key, var):
            if hasattr(var,'items'):
                for k, v in var.items():
                    if k == key:
                        yield v
                    if isinstance(v, dict):
                        for result in gen_dict_extract(key, v):
                            yield result
                    elif isinstance(v, list):
                        for d in v:
                            for result in gen_dict_extract(key, d):
                                yield result

        db              = smdb.SMDB(
                            Namespace(str_logDir = self.arg['dblogbasepath'])
                        )
        db.housingDirs_create()
        d_filteredHits  = self.arg['reportData']

        # In the case of in-line updates on the progress of the
        # postprocess, we need to create a presentation/report object
        # which we can use for the reporting.
        # pudb.set_trace()
        presenter   = report.Report({
                                        'colorize' :    'dark',
                                        'reportData':   d_filteredHits
                                    })
        presenter.run()
        l_run               = []
        d_ret               = {
            'do'        : False
        }
        b_status            : bool = False

        l_then              = self.arg['then'].split(',')
        l_thenArgs          = self.arg['thenArgs'].split(';')
        d_then      : dict  = {}
        d_thenArgs  : dict  = {}
        then        : str   = ""
        if len(l_thenArgs) != len(l_then):
            l_thenArgs      = [''] * len(l_then)
        b_headerPrinted     = False
        thenIndex           = -1
        for (then, thenArgs) in zip(l_then, l_thenArgs):
            if len(thenArgs):
                d_thenArgs  = json.loads(thenArgs)
            else:
                d_thenArgs  = {}
            thenIndex  += 1
            studyIndex  = 0
            d_ret['%02d-%s' % (thenIndex, then)]= { 'study' : []}
            for study in d_filteredHits['data']:
                l_run       = []
                seriesIndex = 0
                if self.arg['withFeedBack']:
                    print("")
                    print(  Colors.BLUE_BCKGRND + Colors.WHITE + "[ STUDY %s ]"\
                            % then + Colors.NO_COLOUR)
                    presenter.studyHeader_print(
                        studyIndex  = studyIndex, reportType = 'rawText'
                    )
                    print(  Colors.BLUE_BCKGRND + Colors.WHITE + "[ SERIES %s ]"\
                            % then + Colors.NO_COLOUR)
                for series in study['series']:
                    str_seriesDescription           = series['SeriesDescription']['value']
                    str_seriesUID                   = series['SeriesInstanceUID']['value']
                    str_studyUID                    = study['StudyInstanceUID']['value']
                    db.d_DICOM['SeriesInstanceUID'] = str_seriesUID
                    self.arg['SeriesInstanceUID']   = str_seriesUID
                    self.arg['StudyInstanceUID']    = str_studyUID
                    if then == "retrieve":  d_then  = retrieve_do()
                    if then == "status"  :  d_then  = status_do()
                    if then == "push"    :  d_then  = push_do(d_thenArgs)
                    if then == "register":  d_then  = register_do(d_thenArgs)
                    if then == "report" :   d_then  = report_do()
                    l_run.append(d_then)
                    if 'status' in d_then:
                        if not d_then['status'] and then != "status": break
                        if d_then['status'] == 'error':
                            if 'withFeedBack' in self.arg:
                                if self.arg['withFeedBack']: print(json.dumps(d_then, indent=4))
                    seriesIndex += 1
                d_ret['%02d-%s' % (thenIndex, then)]['study'].append(
                                { study['StudyInstanceUID']['value'] : l_run}
                )
                studyIndex += 1
                d_ret['do'] = True
        return d_ret

    def xinetd_command(self, opt={}):
        return "service xinetd restart"

    def movescu_command(self, opt={}):
        command = '-S --move ' + opt['aet']
        command += ' --timeout 5'
        command += ' -k QueryRetrieveLevel=SERIES'
        command += ' -k SeriesInstanceUID=' + opt['series_uid']
        command += ' -k StudyInstanceUID='  + opt['study_uid']

        str_cmd     = "%s %s %s" % (
                        self.movescu,
                        command,
                        self.commandSuffix()
        )
        return str_cmd




# Turn off all logging for modules in this libary!!
# Any log noise from pydicom will BREAK receiving
# DICOM data from the remote PACS since the log messages
# will pollute and destroy the DICOM storescp protocol.
import logging
logging.disable(logging.CRITICAL)

# Global modules
import  os
import  subprocess
import  uuid
import  shutil
import  configparser
import  json
from    pathlib         import Path
# PyDicom module
import  pydicom

# PYPX modules
import  pypx.utils

import  pudb
from    pudb.remote         import set_trace

import  pfmisc

class Listen():
    """docstring for Listen."""
    def __init__(self, args):

        self.__name__           = 'Listen'

        self.tmp_directory      = args['tmp_directory']
        self.log_directory      = args['log_directory']
        self.data_directory     = args['data_directory']
        self.executable         = args['executable']

        # Debugging control
        self.b_useDebug         = True
        self.str_debugFile      = '%s/listen.run.log' % self.log_directory
        self.b_quiet            = True
        self.dp                 = pfmisc.debug(
                                            verbosity   = 0,
                                            level       = -1,
                                            within      = self.__name__,
                                            debugToFile = self.b_useDebug,
                                            debugFile   = self.str_debugFile
                                            )

        self.dp.qprint('───────────────────────────────────────────────────────────────────────────', level = -1)
        self.dp.qprint('Incoming DICOM data arriving...', level = -1)

        # maybe it should not create it, as it is a requirement
        os.makedirs(self.tmp_directory,     exist_ok=True)
        os.makedirs(self.log_directory,     exist_ok=True)
        os.makedirs(self.data_directory,    exist_ok=True)

        self.series_mapDir  = os.path.join(self.log_directory, 'series_map')
        self.study_mapDir   = os.path.join(self.log_directory, 'study_map')
        self.patient_mapDir = os.path.join(self.log_directory, 'patient_map')
        os.makedirs(self.series_mapDir,     exist_ok=True)
        os.makedirs(self.study_mapDir,      exist_ok=True)
        os.makedirs(self.patient_mapDir,    exist_ok=True)

        # create unique directory to store inconming data
        self.uuid           = str(uuid.uuid4())
        self.uuid_directory = os.path.join(self.tmp_directory, self.uuid)
        self.log_error      = os.path.join(self.log_directory,
                                            'err-' + self.uuid + '.txt')
        self.log_output     = os.path.join(self.log_directory,
                                            'out-' + self.uuid + '.txt')

        self.dp.qprint('Creating temp holding dir:  %s' % self.uuid_directory,
                        level = -1)
        self.mkdir(self.uuid_directory, self.log_error)

    def mkdir(self, path, log_file):

        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError as e:
                errorfile = open(log_file, 'w')
                errorfile.write('Make ' + path + ' directory\n')
                errorfile.write('Error number: ' + str(e.errno) + '\n')
                errorfile.write('File name: ' + e.filename + '\n')
                errorfile.write('Error message: ' + e.strerror + '\n')
                errorfile.close()

        if not os.path.exists(path):
            errorfile = open(log_file, 'w')
            errorfile.write('File doesn\'t exist:' + path + '\n')
            errorfile.close()
            raise NameError('File doesn\'t exist:' + path)

    def saveInformation(self, path, info):

        if not os.path.exists(path):

            with open(path, 'w') as info_file:
                try:
                    info.write(info_file)
                except OSError as e:
                    errorfile = open(self.log_error, 'w')
                    errorfile.write('Write ' + path + ' file\n')
                    errorfile.write('Error number: ' + str(e.errno) + '\n')
                    errorfile.write('File name: ' + e.filename + '\n')
                    errorfile.write('Error message: ' + e.strerror + '\n')
                    errorfile.close()

        if not os.path.exists(path):
            errorfile = open(self.log_error, 'w')
            errorfile.write('File doesn\'t exist:' + path + '\n')
            errorfile.close()
            raise NameError('File doesn\'t exist:' + path)

    def processDicomField(self, dcm_info, field):
        value = "no value provided"
        if field in dcm_info:
            value = pypx.utils.sanitize(dcm_info.data_element(field).value)
        return value

    def processPatient(self, dcm_info, log_file, data_directory):
        # get information of interest
        patient_id      = self.processDicomField(dcm_info, "PatientID")
        patient_name    = self.processDicomField(dcm_info, "PatientName")
        # self.dp.qprint('Processing PatientID %s...' % patient_id, level = -1)

        # log it
        log_file.write('    PatientID: ' + patient_id + '\n')
        log_file.write('    PatientName: ' + patient_name + '\n')

        # create patient directory
        patient_directory = pypx.utils.patientPath(
                                data_directory,
                                patient_id,
                                patient_name)
        self.mkdir(patient_directory, self.log_error)

        # Save a mapping from this series_uid to the acutal FS location
        # where the file will be written
        str_mapFile     = os.path.join(self.patient_mapDir, '%s.json' % patient_id)
        if not os.path.exists(str_mapFile):
            self.mapFile_save( {patient_id : patient_directory}, str_mapFile )

        # create patient.info file
        patient_info = configparser.ConfigParser()
        patient_info['PATIENT'] = {
            'PatientID': patient_id,
            'PatientName': patient_name,
            'Location': patient_directory
        }

        return patient_info

    def processStudy(self, dcm_info, log_file, patient_directory):
        # get information of interest
        study_description   = self.processDicomField(dcm_info, "StudyDescription")
        study_date          = self.processDicomField(dcm_info, "StudyDate")
        study_uid           = self.processDicomField(dcm_info, "StudyInstanceUID")
        # self.dp.qprint('Processing Study %s...' % study_uid, level = -1)

        # log it
        log_file.write('      StudyDescription: ' + study_description + '\n')
        log_file.write('      StudyDate: ' + study_date + '\n')
        log_file.write('      StudyInstanceUID: ' + study_uid + '\n')

        # create study directory
        study_directory = pypx.utils.studyPath(
            patient_directory, study_description, study_date, study_uid)
        self.mkdir(study_directory, self.log_error)

        # Save a mapping from this series_uid to the acutal FS location
        # where the file will be written
        str_mapFile     = os.path.join(self.study_mapDir, '%s.json' % study_uid)
        if not os.path.exists(str_mapFile):
            self.mapFile_save( {study_uid : study_directory}, str_mapFile )

        # create study.info file
        study_info = configparser.ConfigParser()
        study_info['STUDY'] = {
            'StudyDescription': study_description,
            'StudyDate': study_date,
            'StudyInstanceUID': study_uid,
            'Location': study_directory
        }

        return study_info

    def processSeries(self, dcm_info, log_file, study_directory):
        # get information of interest
        series_description  = self.processDicomField(dcm_info, "SeriesDescription")
        series_date         = self.processDicomField(dcm_info, "SeriesDate")
        series_uid          = self.processDicomField(dcm_info, "SeriesInstanceUID")
        # self.dp.qprint('Processing Series %s' % series_uid, level = -1)

        # log it
        log_file.write('        SeriesDescription: ' + series_description + '\n')
        log_file.write('        SeriesDate: ' + series_date + '\n')
        log_file.write('        SeriesInstanceUID: ' + series_uid + '\n')

        # create series directory
        series_directory = pypx.utils.seriesPath(
            study_directory, series_description, series_date, series_uid)
        self.mkdir(series_directory, self.log_error)

        # Save a mapping from this series_uid to the acutal FS location
        # where the file will be written
        str_mapFile     = os.path.join(self.series_mapDir, '%s.json' % series_uid)
        if not os.path.exists(str_mapFile):
            self.mapFile_save( {series_uid : series_directory}, str_mapFile )

        # store information as a configuration
        series_info = configparser.ConfigParser()
        series_info['SERIES'] = {
            'SeriesDescription': series_description,
            'SeriesDate': series_date,
            'SeriesInstanceUID': series_uid,
            'Location': series_directory
        }

        return series_info

    def mapFile_save(self, ad_json, astr_mapFile):
        """
        Save a dictionary <ad_json> in <astr_mapDir>/<astr_mapFile>
        """
        b_ret           = False

        if not os.path.exists(astr_mapFile):
            try:
                with open(astr_mapFile, 'w') as f:
                    json.dump(ad_json, f)
                f.close()
                b_ret   = True
            except:
                b_ret   = False
        return b_ret

    def processImage(self, dcm_info, log_file, error_file, series_directory, tmp_file):
        # get information of interest
        image_uid               = self.processDicomField(dcm_info, "SOPInstanceUID")
        image_instance_number   = self.processDicomField(dcm_info, "InstanceNumber")

        # log it
        log_file.write('          SOPInstanceUID: ' + image_uid + '\n')
        log_file.write('          InstanceNumber: ' + image_instance_number + '\n')

        image_path = pypx.utils.dataPath(series_directory, image_instance_number, image_uid)

        if not os.path.exists(image_path):
            try:
                shutil.copy2(tmp_file, image_path)
            except OSError as e:
                errorfile = open(error_file, 'w')
                errorfile.write('Copy ' + tmp_file + ' to ' + image_path + '\n')
                errorfile.write('Error number: ' + str(e.errno) + '\n')
                errorfile.write('File name: ' + e.filename + '\n')
                errorfile.write('Error message: ' + e.strerror + '\n')
                errorfile.close()

        if not os.path.exists(image_path):
            errorfile = open(error_file, 'w')
            errorfile.write('File doesn\'t exist:' + image_path + '\n')
            errorfile.close()
            raise NameError('File doesn\'t exist:' + image_path)

    def run(self):

        def transmission_summarise():
            """
            Log a summary of transmission data to listen log file.
            """

            study_description   = self.processDicomField(dcm_info, "StudyDescription")
            study_date          = self.processDicomField(dcm_info, "StudyDate")
            series_description  = self.processDicomField(dcm_info, "SeriesDescription")
            patient_id          = self.processDicomField(dcm_info, "PatientID")
            patient_name        = self.processDicomField(dcm_info, "PatientName")
            protocol_name       = self.processDicomField(dcm_info, "ProtocolName")
            d_fileInfo          = filesInSeries_determine()
            self.dp.qprint('Summary report:')
            self.dp.qprint('PatientID:                  %s' % patient_id, level = -1)
            self.dp.qprint('PatientName:                %s' % patient_name, level = -1)
            self.dp.qprint('StudyDate:                  %s' % study_date, level = -1)
            self.dp.qprint('StudyDescription:           %s' % study_description, level = -1)
            self.dp.qprint('SeriesDescription:          %s' % series_description, level = -1)
            self.dp.qprint('ProtocolName:               %s' % protocol_name, level = -1)
            if d_fileInfo['status']:
                self.dp.qprint('Number of files in Series:  %d' % d_fileInfo['fileCount'], level = -1)
                self.dp.qprint('Directory size (raw):       %d' % d_fileInfo['dirSizeRaw'], level = -1)
                self.dp.qprint('Directory size (human):     %s' % d_fileInfo['str_dirSize'], level = -1)

        def filesInSeries_determine():
            """
            Determine the number of files in a given series and some
            dir info
            """
            def du(path):
                """disk usage in human readable format (e.g. '2,1GB')"""
                return subprocess.check_output(['du','-sh', path]).split()[0].decode('utf-8')

            def duRaw(path):
                root    = Path(path)
                return  sum(f.stat().st_size for f in root.glob('**/*') if f.is_file())

            series_uid          = self.processDicomField(dcm_info, "SeriesInstanceUID")
            str_seriesMapFile   = os.path.join(self.series_mapDir, '%s.json' % series_uid)

            try:
                with open(str_seriesMapFile, 'r') as f:
                    d_seriesInfo    = json.load(f)
                str_path            = d_seriesInfo[series_uid]
                fileCount           = len([n for n in os.listdir(str_path) \
                                        if os.path.isfile(os.path.join(str_path, n))])
                str_dirSize         = du(str_path)
                dirSizeRaw          = duRaw(str_path)
                d_ret               = {
                    'status':       True,
                    'fileCount':    fileCount,
                    'str_dirSize':  str_dirSize,
                    'dirSizeRaw':   dirSizeRaw
                }
            except:
                d_ret               = {
                    'status':       False,
                    'fileCount':    -1,
                    'str_dirSize':  "unknown",
                    'dirSizeRaw':   -1
                }

            return d_ret

        # start listening to incoming data
        self.dp.qprint("Listening for and parsing incoming data...", level = -1)

        # command = self.executable   + ' -id -od "'              + \
        #   self.uuid_directory       + '" -xcr "touch '          + \
        #   self.uuid_directory       + '/#f;touch '              + \
        #   self.uuid_directory       + '/#c;touch '              + \
        #   self.uuid_directory       + '/#a" -pm -sp -d -lc '    + \
        #   self.uuid_directory       + '/debug.log'

        command = self.executable   + ' -id -od '               + \
          self.uuid_directory       + ' -pm -sp -d >'        + \
          self.uuid_directory       + '/debug.log'

        self.dp.qprint("Listener command:... \n'%s'" % command, level = -1)
        subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)

        abs_files = [os.path.join(self.uuid_directory,f) for f in os.listdir(self.uuid_directory)]
        abs_dirs = [f for f in list(abs_files) if os.path.isdir(f)]

        # Start logging
        stdout_file = open(self.log_output, 'w')
        stdout_file.write('UUID DIRECTORY:' + '\n')
        stdout_file.write(self.uuid_directory + '\n')

        # Keep track of "receiving.series" files
        series_received = set()

        for directory in abs_dirs:

            stdout_file.write('> ' + directory + '\n')
            for data in os.listdir(directory):

                stdout_file.write('>>> ' + data + '\n')
                # abs path to data
                abs_data = os.path.join(directory, data)
                dcm_info = pydicom.dcmread(abs_data)

                # process patient
                patient_info = self.processPatient(dcm_info, stdout_file, self.data_directory)
                patient_info_path = os.path.join(
                    patient_info['PATIENT']['Location'], 'patient.info')
                self.saveInformation(patient_info_path, patient_info)

                # process study
                study_info = self.processStudy(
                    dcm_info, stdout_file, patient_info['PATIENT']['Location'])
                study_info_path = os.path.join(study_info['STUDY']['Location'], 'study.info')
                self.saveInformation(study_info_path, study_info)

                # process series
                series_info = self.processSeries(
                    dcm_info, stdout_file, study_info['STUDY']['Location'])
                series_info_path = os.path.join(
                    series_info['SERIES']['Location'], 'receiving.series')
                self.saveInformation(series_info_path, series_info)
                # keep track of series we are receiving
                # receiving.series will be rename to series.info
                # when all the series has been received
                series_received.add(series_info_path)

                # process image
                self.processImage(
                    dcm_info, stdout_file, self.log_error,
                    series_info['SERIES']['Location'], abs_data)
                # image.info file
                # mri_info? :/

        # rename receiving.series to series.info
        # changing name lets external applications know the incoming data has been received
        for series in series_received:
            last_index = series.rfind('receiving.series')
            target = series[:last_index] + 'series.info'
            os.rename(series, target)

        self.dp.qprint('All images processed.', level = -1)
        transmission_summarise()

        # cleanup
        # try:
        #     shutil.rmtree(self.uuid_directory)
        #     os.remove(self.log_output)
        # except OSError as err:
        #     errorfile = open(self.log_error, 'w')
        #     errorfile.write('Remove ' + self.uuid_directory + ' tree\n')
        #     errorfile.write('Error number: ' + str(err.errno) + '\n')
        #     errorfile.write('File name: ' + err.filename + '\n')
        #     errorfile.write('Error message: ' + err.strerror + '\n')
        #     errorfile.close()

        # what about log files?
        # import logger?

        self.dp.qprint('DICOM transmission complete.', level = -1)

        stdout_file.close()



# Global modules
import  subprocess
import  pudb
import  json
import  pfmisc
from    pfmisc._colors      import  Colors

# PYPX modules
from .base import Base

class Move(Base):
    """
    The 'Move' class is essentially a stripped down module that
    simply performs a call to the system to run an appropriately
    constructed 'movescu' command.

    In some ways, the Move() class replicates/duplicates similar
    functionality in the Find() class. Future development might
    address this overlap more intelligently.
    """

    def __init__(self, arg):
        """
        Constructor.

        Largely simple/barebones constructor that calls the Base()
        and sets up the executable name.
        """

        super(Move, self).__init__(arg)
        self.dp     = pfmisc.debug(
                        verbosity   = self.verbosity,
                        within      = 'Find',
                        syslog      = False
        )
        self.log    = self.dp.qprint

    def movescu_command(self, opt={}):
        command = '-S --move ' + opt['aet']
        command += ' --timeout 5'
        command += ' -k QueryRetrieveLevel=SERIES'
        command += ' -k StudyInstanceUID='  + opt['study_uid']
        command += ' -k SeriesInstanceUID=' + opt['series_uid']

        str_cmd     = "%s %s %s" % (
                        self.movescu,
                        command,
                        self.commandSuffix()
        )
        return str_cmd

    def run(self, opt={}):

        # First, for dockerized run, (re)start the xinetd service
        # self.systemlevel_run(self.arg,
        #     {
        #         'executable':   'xinetd'
        #     }
        # )

        d_moveRun = self.systemlevel_run(self.arg,
                {
                    'f_commandGen':         self.movescu_command,
                    'series_uid':           opt['SeriesInstanceUID'],
                    'study_uid':            opt['StudyInstanceUID']
                }
        )

        return d_moveRun



# Turn off all logging for modules in this libary!!
# Any log noise from pydicom will BREAK receiving
# DICOM data from the remote PACS since the log messages
# will pollute and destroy the DICOM storescp protocol.
import logging
logging.disable(logging.CRITICAL)

from    argparse            import  Namespace

# Global modules
import  os
import  sys
from    os                  import  listdir
from    os.path             import  isfile, join
import  subprocess
import  uuid
import  shutil
import  configparser
import  json
from    pathlib             import  Path
import  uuid
import  pathlib
import  datetime
from    datetime            import  date, datetime
import  inspect
import  hashlib
import  re

# PyDicom module
import  pydicom             as      dicom

# PYPX modules
import  pypx.utils
import  pypx.smdb

# Debugging
import  pudb
from    pudb.remote         import  set_trace
import  pfmisc

from    argparse            import  Namespace, ArgumentParser
from    argparse            import  RawTextHelpFormatter

def parser_setup(str_desc):
    parser = ArgumentParser(
                description         = str_desc,
                formatter_class     = RawTextHelpFormatter
            )

    # JSONarg
    parser.add_argument(
        '--JSONargs',
        action  = 'store',
        dest    = 'JSONargString',
        type    = str,
        default = '',
        help    = 'JSON equivalent of CLI key/values')

    parser.add_argument(
        '-p', '--xcrdir',
        action  = 'store',
        dest    = 'str_xcrdir',
        type    = str,
        default = '/tmp',
        help    = 'Directory containing a received study'
        )
    parser.add_argument(
        '-f', '--xcrfile',
        action  = 'store',
        dest    = 'str_xcrfile',
        type    = str,
        default = '',
        help    = 'File in <xcrdir> to process'
        )
    parser.add_argument(
        '--xcrdirfile',
        action  = 'store',
        dest    = 'str_xcrdirfile',
        type    = str,
        default = '',
        help    = 'Fully qualified file to process'
        )
    parser.add_argument(
        '--parseAllFilesWithSubStr',
        action  = 'store',
        dest    = 'str_filesubstr',
        type    = str,
        default = '',
        help    = 'Parse all files in <xcrdir> that contain <substr>'
        )
    parser.add_argument(
        '-l', '--logdir',
        action  = 'store',
        dest    = 'str_logDir',
        type    = str,
        default = '/tmp/log',
        help    = 'Directory to store log files'
        )
    parser.add_argument(
        '-d', '--datadir',
        action  = 'store',
        dest    = 'str_dataDir',
        type    = str,
        default = '/tmp/data',
        help    = 'Directory in which to pack final DICOM files'
        )

    parser.add_argument(
        '--rootDirTemplate',
        action  = 'store',
        dest    = 'str_rootDirTemplate',
        type    = str,
        default = '%PatientID-%PatientName-%PatientBirthDate',
        help    = 'Template pattern for root unpack directory'
        )
    parser.add_argument(
        '--studyDirTemplate',
        action  = 'store',
        dest    = 'str_studyDirTemplate',
        type    = str,
        default = '%StudyDescription-%AccessionNumber-%StudyDate',
        help    = 'Template pattern for study unpack directory'
        )
    parser.add_argument(
        '--seriesDirTemplate',
        action  = 'store',
        dest    = 'str_seriesDirTemplate',
        type    = str,
        default = '%_pad|5,0_SeriesNumber-%SeriesDescription',
        help    = 'Template pattern for series unpack directory'
        )
    parser.add_argument(
        '--imageTemplate',
        action  = 'store',
        dest    = 'str_imageTemplate',
        type    = str,
        default = '%_pad|4,0_InstanceNumber-%SOPInstanceUID.dcm',
        help    = 'Template pattern for image file'
        )

    parser.add_argument(
        '--cleanup',
        action  = 'store_true',
        dest    = 'b_cleanup',
        default = False,
        help    = 'If specified, then cleanup temporary files'
        )
    parser.add_argument(
        '--debug',
        action  = 'store_true',
        dest    = 'b_debug',
        default = False,
        help    = 'If specified, then also log debug info to <logdir>'
        )
    parser.add_argument(
        "-v", "--verbosity",
        help    = "verbosity level for app",
        dest    = 'verbosity',
        type    = int,
        default = 1)
    parser.add_argument(
        "-x", "--desc",
        help    = "show long synopsis",
        dest    = 'b_desc',
        action  = 'store_true',
        default = False
    )
    parser.add_argument(
        "-y", "--synopsis",
        help    = "show short synopsis",
        dest    = 'b_synopsis',
        action  = 'store_true',
        default = False
    )
    parser.add_argument(
        '--version',
        help    = 'if specified, print version number',
        dest    = 'b_version',
        action  = 'store_true',
        default = False
    )
    return parser

def parser_interpret(parser, *args):
    """
    Interpret the list space of *args, or sys.argv[1:] if
    *args is empty
    """
    if len(args):
        args, unknown    = parser.parse_known_args(*args)
    else:
        args, unknown    = parser.parse_known_args(sys.argv[1:])
    return args, unknown

def parser_JSONinterpret(parser, d_JSONargs):
    """
    Interpret a JSON dictionary in lieu of CLI.

    For each <key>:<value> in the d_JSONargs, append to
    list two strings ["--<key>", "<value>"] and then
    argparse.
    """
    l_args  = []
    for k, v in d_JSONargs.items():
        l_args.append('--%s' % k)
        if type(v) == type(True): continue
        l_args.append('%s' % v)
    return parser_interpret(parser, l_args)


def args_impedanceMatch(ns_arg):
    """
    This method is an "impedance matcher" that examines the
    incoming namespace, ns_arg, and returns a new namespace that
    contains any missing elements necessary for full instantiation
    of the class object.

    Typically this method is used when the class is called as a module
    without the assumption of the native driving script creating the
    the fully qualified namespace.
    """
    l_key   : list  = []

    # Get the parser structure for this module
    parser          = parser_setup("impedanceMatching")
    args, unknown   = parser_interpret(parser)

    l_key = [k for (k,v) in vars(ns_arg).items()]
    if 'str_xcrdir'     not in l_key:   setattr(ns_arg, 'str_xcrdir', '/tmp')
    if 'str_xcrfile'    not in l_key:   setattr(ns_arg, 'str_xcrfile', '')
    if 'str_xcrdir'     not in l_key:   setattr(ns_arg, 'str_xcrdir', '')
    if 'str_xcrdirfile' not in l_key:   setattr(ns_arg, 'str_xcrdirfile', '')
    if 'str_filesubstr' not in l_key:   setattr(ns_arg, 'str_filesubstr', '')

    if 'str_rootDirTemplate'    not in l_key:
        setattr(ns_arg, 'str_rootDirTemplate',      args.str_rootDirTemplate)
    if 'str_studyDirTemplate'   not in l_key:
        setattr(ns_arg, 'str_studyDirTemplate',     args.str_studyDirTemplate)
    if 'str_seriesDirTemplate'  not in l_key:
        setattr(ns_arg, 'str_seriesDirTemplate',    args.str_seriesDirTemplate)
    if 'str_imageTemplate'      not in l_key:
        setattr(ns_arg, 'str_imageTemplate',        args.str_imageTemplate)
    return ns_arg

class Process():
    """
    The core class of the repack module -- this class essentially reads
    a DICOM file, parses its tags, and then repacks (or re-copies) that
    file to a more descriptive location on the filesystem.
    """

    def loggers_create(self):
        """
        >>>>>>>>>>>>>>   Debugging control  <<<<<<<<<<<<<<<<
        Essentially we create some pfmisc.debug objects that
        write to files and also give them some shortcut names
        """
        str_thisMRsession       = pathlib.PurePath(self.args.str_xcrdir).name
        self.str_pulseFile      = '%s/%s-pulse.log'     % (
                                                            self.args.str_logDir,
                                                            str_thisMRsession
                                                        )
        self.str_debugFile      = '%s/repack.log'       % self.args.str_logDir
        self.pulseObj           = pfmisc.debug(
                                            verbosity   = int(self.args.verbosity),
                                            level       = 2,
                                            within      = self.__name__,
                                            debugToFile = self.args.b_debug,
                                            debugFile   = self.str_pulseFile
                                            )
        self.dp                 = pfmisc.debug(
                                            verbosity   = int(self.args.verbosity),
                                            level       = 2,
                                            within      = self.__name__,
                                            debugToFile = self.args.b_debug,
                                            debugFile   = self.str_debugFile
                                            )
        self.logPulse           = self.pulseObj.qprint
        self.log                = self.dp.qprint

    def filesToRepack_determine(self):
        """
        Based on the pattern of CLI calling flags, determine which
        files to repack.
        """
        if self.args.str_filesubstr:
            try:
                # First create a list of all the files...
                self.l_files    = [
                    f                                                       \
                        for f in listdir(self.args.str_xcrdir)              \
                            if isfile(join(self.args.str_xcrdir, f))
                ]
                # Now filter them according to the passed filesubstr
                self.l_files    = [
                    x                                                       \
                        for y in self.args.str_filesubstr.split(',')        \
                            for x in self.l_files if y in x
                ]
            except:
                pass
        else:
            self.l_files.append(self.args.str_xcrfile)

    def __init__(self, args):

        self.__name__           : str   = 'repack'
        self.args                       = args
        self.l_files            : list  = []

        if len(self.args.str_xcrdirfile):
            self.args.str_xcrdir        = os.path.dirname(
                                                self.args.str_xcrdirfile
                                        )
            self.args.str_xcrfile       = os.path.basename(
                                                self.args.str_xcrdirfile
                                        )

        self.smdb                       = pypx.smdb.SMDB(args)

        # set_trace(
        #             host            = "0.0.0.0",
        #             port            = 5555,
        #             term_size       = (252, 63)
        #         )

        self.filesToRepack_determine()
        self.loggers_create()
        self.logPulse('Pulsing on dir %s...' %  self.args.str_xcrdir,
                      level = 2)
        self.log(
            'Incoming DICOM dir: %s' % (self.args.str_xcrdir),
            level = 2
        )
        for str_file in self.l_files:
            self.log(
                'Incoming DICOM file: %s' % (str_file),
                level = 2
            )

    def cleanup(self, str_file):
        """
        Clean up -- this is called once per process sweep, with the
        received <str_file> as argument. Quite simply this method just
        deletes that file, should it exist, in the <xcrdir>.
        """
        b_status        :   bool    = False
        str_filepath    :   str     = os.path.join(
                                        self.args.str_xcrdir,
                                        str_file
                                    )
        str_error       :   str     = ''
        str_message     :   str     = ''
        if os.path.isfile(str_filepath):
            try:
                os.remove(str_filepath)
                b_status    = True
                str_message = '%s successfully deleted' % str_filepath
            except Exception as e:
                str_error   = '%s' % e

        return {
            'status'    : b_status,
            'error'     : str_error,
            'message'   : str_message
        }

    def run(self) -> dict:
        """
        Main entry point for receiver. This will, for each DICOM file to
        process, read the DICOM, perform some preprocessing on the DICOM
        tag  space,  and then  save  the  file  in the appropriate target
        directory tree.
        """
        dl_run      : list  = []
        d_run       : dict  = {'status' : False}
        for str_file in self.l_files:
            d_run           = self.DICOMfile_mapsUpdate(
                                self.DICOMfile_save(
                                    Process.DICOMfile_read(
                                        file = '%s/%s' % (
                                                self.args.str_xcrdir,
                                                str_file
                                            )
                                    )
                                )
                            )
            self.log(
                'DICOM repacked to: %s/%s' % \
                    ( d_run['d_DICOMfile_save']['outputDir'],
                      d_run['d_DICOMfile_save']['outputFile']),
                level = 2
            )
            # Before returning, we need to "sanitize" some of the
            # DICOMfile_read fields that are not JSON serializable
            # allowing us to present the caller with a nice return
            # JSON payload
            d_DICOM             = d_run['d_DICOMfile_save']\
                                            ['d_DICOMfile_read']\
                                                ['d_DICOM']
            d_DICOM['dcm']      = "Not JSON serializable"
            d_DICOM['d_dcm']    = "Not JSON serializable"
            d_DICOM['d_dicom']  = "Not JSON serializable"
            if self.args.b_cleanup: d_run['cleanup']= self.cleanup(str_file)
            dl_run.append(d_run)

        return {
            'status'    : d_run['status'],
            'run'       : dl_run
        }

    def packPath_resolve(self, d_DICOMfile_read) -> dict:
        """
        Return the pack path and image name template. Note this
        needs a d_DICOMread dictionary as returned from a call
        to DICOMfile_read.
        """

        def DICOMlookup_santitizeFromTemplate(str_template):
            """
            Process DICOM lookup tags in a template string and
            return a sanitized result.
            """
            return re.sub('[^A-Za-z0-9\.\-]+', '_',
                                self.tagsInString_process(
                                    d_DICOMfile_read['d_DICOM'],
                                    str_template
                                )['str_result']
                            )

        str_rootDir     :   str     = ''
        str_studyDir    :   str     = ''
        str_seriesDir   :   str     = ''
        str_packDir     :   str     = ''
        str_imageFile   :   str     = ''

        str_rootDir     = DICOMlookup_santitizeFromTemplate(
                            self.args.str_rootDirTemplate
                        )
        str_studyDir    = DICOMlookup_santitizeFromTemplate(
                            self.args.str_studyDirTemplate
                        )
        str_seriesDir   = DICOMlookup_santitizeFromTemplate(
                            self.args.str_seriesDirTemplate
                        )
        str_imageFile   = DICOMlookup_santitizeFromTemplate(
                            self.args.str_imageTemplate
                        )
        str_packDir     = '%s/%s/%s' % (
                            str_rootDir,
                            str_studyDir,
                            str_seriesDir
                        )

        return {
            'status'    : True,
            'packDir'   : str_packDir,
            'imageFile' : str_imageFile
        }


    def DICOMfile_save(self, d_DICOMfile_read) -> dict:
        """
        Save/pack the initial DICOM file in a new location based on the
        various <template> patterns:

        <dataDir>
            |
            └─<rootTemplate>
                    |
                    └─<studyTemplate>
                             |
                             └─<seriesTemplate>
                                      |
                                      └─<imageTemplate>.dcm

        Also update the various map files that are used to track the
        status of receipts.
        """

        b_status        :   bool    = False
        str_imageFile   :   str     = ''
        str_outputDir   :   str     = ''
        str_errorDir    :   str     = ''
        str_errorCopy   :   str     = ''
        str_path        :   str     = ''
        d_path          :   dict    = {}

        # pudb.set_trace()
        if d_DICOMfile_read['status']:
            d_path          = self.packPath_resolve(d_DICOMfile_read)
            str_outputDir   = '%s/%s' % (
                                    self.args.str_dataDir,
                                    d_path['packDir']
                            )
            str_imageFile   = d_path['imageFile']

            try:
                os.makedirs(str_outputDir)
            except Exception as e:
                str_errorDir    = '%s' %e
            try:
                str_path    = shutil.copy(
                                '%s/%s' % (
                                    d_DICOMfile_read['inputPath'],
                                    d_DICOMfile_read['inputFileName']
                                ),
                                '%s/%s' % (
                                    str_outputDir,
                                    str_imageFile
                                )
                            )
                b_status    = True
            except Exception as e:
                str_errorCopy   = '%s' % e
        return {
            'method'            : inspect.stack()[0][3],
            'outputDir'         : str_outputDir,
            'outputFile'        : str_imageFile,
            'shutilpath'        : str_path,
            'status'            : b_status,
            'errorDir'          : str_errorDir,
            'errorCopy'         : str_errorCopy,
            'd_DICOMfile_read'  : d_DICOMfile_read
        }

    def DICOMfile_mapsUpdate(self, d_DICOMfile_save)    -> dict:
        """
        Interact with the SMDB object to update JSON mapping information
        relative to this save operation.
        """
        b_status        :   bool    = False
        d_mapsUpdate    :   dict    = {}

        if d_DICOMfile_save['status']:
            b_status        = True
            self.smdb.housingDirs_create()
            self.smdb.DICOMobj_set(d_DICOMfile_save ['d_DICOMfile_read']\
                                                    ['d_DICOM']\
                                                    ['d_dicomSimple'])
            d_mapsUpdate    = self.smdb.mapsUpdateForFile(
                    '%s/%s' % ( d_DICOMfile_save['outputDir'],
                                d_DICOMfile_save['outputFile'])
            )
            self.smdb.seriesData('pack', 'seriesPack', True)

        return {
            'status'            : b_status,
            'd_mapsUpdate'      : d_mapsUpdate,
            'd_DICOMfile_save'  : d_DICOMfile_save
        }

    @staticmethod
    def DICOMfile_read(*args, **kwargs) -> dict:
        """
        Read a DICOM file and perform some initial  parsing of tags,
        returning  a  dictionary  object of multiple representations
        of the dicom data.

        Nested functions are used here, mainly for encapsulation
        and readability.
        """

        # Core structure template returned by this method
        d_DICOM : dict = {
                'str_dicomFile'     : '',
                'dcm'               : None,
                'd_dcm'             : {},
                'str_raw'           : '',
                'l_tagRaw'          : [],
                'str_json'          : {},
                'd_dicom'           : {},
                'd_dicomSimple'     : {},
                'l_tagsUsed'        : []
        }

        def dcm_readFromFile(str_file, d_DICOM) -> bool:
            """
            Load the <str_file> and populate some field records in <d_DCIOM>.
            """
            b_status    : bool          = False
            d_err       : dict          = {
                                            'file'      : '',
                                            'cwd'       : '',
                                            'message'   : ''
                                        }
            d_DICOM['str_dicomFile']    = str_file
            try:
                d_DICOM['dcm']          = dicom.read_file(str_file)
                b_status                = True
            except Exception as e:
                d_err['file']           = str_file
                d_err['cwd']            = os.getcwd()
                d_err['message']        = '%s' % e
                b_status                = False
            return {
                'method'        : inspect.stack()[0][3],
                'status'        : b_status,
                'error'         : d_err
            }

        def dcm_doExplicitToStr(d_dcm, str_file) -> dict:
            """
            This a nested error mitigation method, called when a
            implicit "to-string"  conversion fails.  This method
            attempts to perform  an  explicit conversion instead
            by performing an element  by element conversion over
            the dictionary of d_dcm FileDataset components.
            """
            b_status    : bool  = False
            l_k         : list  = list(d_dcm.keys())
            str_raw     : str   = ''
            str_err     : str   = ''

            for k in l_k:
                try:
                    str_raw     += str(d_dcm[k])
                    str_raw     += '\n'
                    b_status    = True
                except:
                    str_err     = 'Failed to string convert key "%s"' % k
                    str_raw     += str_err + "\n"
                    b_status    = False
            return {
                'method'        : inspect.stack()[0][3],
                'failingFile'   : str_file,
                'status'        : b_status,
                'conversion'    : str_raw,
                'error'         : str_err
            }

        def dcm_populate(d_DICOM, d_prior)   -> dict:
            """
            Populate some additional records of the d_DICOM structure
            by attempting simple conversions on the pydicom object
            """
            b_status    : bool          = False
            d_explicit  : dict          = {}
            d_raw       : dict          = {}
            if d_prior['status']:
                b_status                = True
                d_DICOM['l_tagRaw']     = d_DICOM['dcm'].dir()
                d_DICOM['d_dcm']        = dict(d_DICOM['dcm'])
                try:
                    d_DICOM['str_raw']  = str(d_DICOM['dcm'])
                except:
                    d_raw               = dcm_doExplicitToStr(
                                            d_DICOM['d_dcm'],
                                            d_prior)
                    d_DICOM['str_raw']  = d_raw['conversion']
            return {
                'method'        : inspect.stack()[0][3],
                'status'        : b_status,
                'rawConversion' : d_raw,
                'prior'         : d_prior
            }

        def dcm_dicomDictsProcess(d_DICOM, l_tags, d_prior) -> dict:
            """
            Populate the d_dicom dictionaries in the housing structure.
            """
            b_status    : bool              = False
            str_error   : str               = ''
            if d_prior['status']:
                if len(l_tags):
                    d_DICOM['l_tagsUsed']   = l_tags
                else:
                    d_DICOM['l_tagsUsed']   = d_DICOM['l_tagRaw'].copy()

                if 'PixelData' in d_DICOM['l_tagsUsed']:
                    d_DICOM['l_tagsUsed'].remove('PixelData')

                for key in d_DICOM['l_tagsUsed']:
                    d_DICOM['d_dicom'][key] = d_DICOM['dcm'].data_element(key)
                    try:
                        d_DICOM['d_dicomSimple'][key] = getattr(d_DICOM['dcm'], key)
                    except:
                        d_DICOM['d_dicomSimple'][key] = "no attribute"
                    if  not isinstance(d_DICOM['d_dicomSimple'][key], str)      or \
                        not isinstance(d_DICOM['d_dicomSimple'][key], list)     or \
                        not isinstance(d_DICOM['d_dicomSimple'][key], float)    or \
                        not isinstance(d_DICOM['d_dicomSimple'][key], bool)     or \
                        not isinstance(d_DICOM['d_dicomSimple'][key], int):
                            d_DICOM['d_dicomSimple'][key] = '%s' %  \
                                d_DICOM['d_dicomSimple'][key]
                try:
                    d_DICOM['str_json']     = json.dumps(d_DICOM['d_dicomSimple'])
                    b_status                = True
                except Exception as e:
                    str_error   = '%s' % e

            return {
                'method'        : inspect.stack()[0][3],
                'status'        : b_status,
                'error'         : str_error,
                'prior'         : d_prior
            }

        b_status        : bool  = False
        l_tags          : list  = []
        l_tagsToUse     : list  = []
        d_tagsInString  : dict  = {}
        str_file        : str   = ""
        str_outputFile  : str   = ""

        for k, v in kwargs.items():
            if k == 'file':             str_file    = v
            if k == 'l_tagsToUse':      l_tags      = v

        if len(args):
            l_file      : list  = args[0]
            str_file    : str   = l_file[0]

        d_DICOMprocess  = dcm_dicomDictsProcess(d_DICOM, l_tags,
            dcm_populate(d_DICOM,
                dcm_readFromFile(str_file, d_DICOM)
            )
        )

        return {
            'method'            : inspect.stack()[0][3],
            'status'            : d_DICOMprocess['status'],
            'inputPath'         : os.path.dirname(str_file),
            'inputFileName'     : os.path.basename(str_file),
            'd_DICOM'           : d_DICOM,
            'd_DICOMprocess'    : d_DICOMprocess
        }


    def tagsInString_process(self, d_DICOM, astr, *args, **kwargs):
        """
        This method substitutes DICOM tags that are '%'-tagged
        in a string template with the actual tag lookup.

        For example, an output filename that is specified as the
        following string:

            %PatientAge-%PatientID-output.txt

        will be parsed to

            006Y-4412364-ouptut.txt

        It is also possible to apply certain permutations/functions
        to a tag. For example, a function is identified by an underscore
        prefixed and suffixed string as part of the DICOM tag. If
        found, this function is applied to the tag value. For example,

            %PatientAge-%_md5|4_PatientID-output.txt

        will apply an md5 hash to the PatientID and use the first 4
        characters:

            006Y-7f38-output.txt

        """

        def pad_process(func, str_replace):
            """
            Pad a string in a given width.

            Arg specifier:

                    "<width>,<leadingChar>"

            i.e.

                %_pad|3,0_

            will pad the value in a width 3 with leading '0's.

            """
            nonlocal    astr
            l_funcTag   = []        # a function/tag list
            l_funcTag   = func.split('_')[1:]
            func        = l_funcTag[0]
            char        = '0'
            str_pad     = ''
            l_args      = func.split('|')
            if len(l_args) > 1:
                str_pad     = l_args[1]
            if len(str_pad):
                l_arg   = str_pad.split(',')
                width   = l_arg[0]
                if len(l_arg) > 1:
                    char    = l_arg[1]
                str_replace     = str_replace.rjust(int(width), char)
            astr        = astr.replace('_%s_' % func, '')
            return astr, str_replace

        def md5_process(func, str_replace):
            """
            md5 mangle the <str_replace>.
            """
            nonlocal    astr
            l_funcTag   = []        # a function/tag list
            l_args      = []        # the 'args' of the function
            chars       = ''        # the number of resultant chars from func
                                    # result to use
            str_replace = hashlib.md5(str_replace.encode('utf-8')).hexdigest()
            l_funcTag   = func.split('_')[1:]
            func        = l_funcTag[0]
            l_args      = func.split('|')
            if len(l_args) > 1:
                chars   = l_args[1]
                str_replace     = str_replace[0:int(chars)]
            astr        = astr.replace('_%s_' % func, '')
            return astr, str_replace

        def strmsk_process(func, str_replace):
            """
            string mask
            """
            nonlocal    astr
            l_funcTag   = []        # a function/tag list
            l_funcTag   = func.split('_')[1:]
            func        = l_funcTag[0]
            str_msk     = func.split('|')[1]
            l_n = []
            for i, j in zip(list(str_replace), list(str_msk)):
                if j == '*':    l_n.append(i)
                else:           l_n.append(j)
            str_replace = ''.join(l_n)
            astr        = astr.replace('_%s_' % func, '')
            return astr, str_replace

        def nospc_process(func, str_replace):
            """
            replace spaces in string
            """
            nonlocal    astr
            l_funcTag   = []        # a function/tag list
            l_args      = []        # the 'args' of the function
            l_funcTag   = func.split('_')[1:]
            func        = l_funcTag[0]
            l_args      = func.split('|')
            str_char    = ''
            if len(l_args) > 1:
                str_char = l_args[1]
            # strip out all non-alphnumeric chars and
            # replace with space
            str_replace = re.sub(r'\W+', ' ', str_replace)
            # replace all spaces with str_char
            str_replace = str_char.join(str_replace.split())
            astr        = astr.replace('_%s_' % func, '')
            return astr, str_replace

        def convertToNumber (s):
            return int.from_bytes(s.encode(), 'little')

        def convertFromNumber (n):
            return n.to_bytes(math.ceil(n.bit_length() / 8), 'little').decode()

        def name_process(func, str_replace):
            """
            replace str_replace with a name

            Note this sub-function can take as an argument a DICOM tag, which
            is then used to seed the name caller. This assures that all
            DICOM files belonging to the same series (or that have the same
            DICOM tag value passed as argument) all get the same 'name'.

            NB: If a DICOM tag is passed as an argument, the first character
            of the tag must be lower case to protect parsing of any non-arg
            DICOM tags.
            """
            nonlocal        astr, d_DICOM
            l_funcTag       = []        # a function/tag list
            l_args          = []        # the 'args' of the function
            l_funcTag       = func.split('_')[1:]
            func            = l_funcTag[0]
            l_args          = func.split('|')
            if len(l_args) > 1:
                str_argTag  = l_args[1]
                str_argTag  = re.sub('([a-zA-Z])', lambda x: x.groups()[0].upper(), str_argTag, 1)
                if str_argTag in d_DICOM['d_dicomSimple']:
                    str_seed    = d_DICOM['d_dicomSimple'][str_argTag]
                    randSeed    = convertToNumber(str_seed)
                    Faker.seed(randSeed)
            str_firstLast   = pfdicom.fake.name()
            l_firstLast     = str_firstLast.split()
            str_first       = l_firstLast[0]
            str_last        = l_firstLast[1]
            str_replace     = '%s^%s^ANON' % (str_last.upper(), str_first.upper())
            astr            = astr.replace('_%s_' % func, '')
            return astr, str_replace

        def age_daysToDMY(ageInDays : int) -> str :
            """
            Given an age in days, return a string of D[ay], M[onth], [Y]ear
            of the age closest to either epoch.
            """
            str_age :   str     = "999Y"
            if ageInDays >= 365:
                str_age = '%03dY' % round(ageInDays/365)
            if ageInDays >= 30 and ageInDays < 365:
                str_age = '%03dM' % round(ageInDays/30)
            if ageInDays >=7 and ageInDays < 30:
                str_age = '%03dW' % round(ageInDays/7)
            if ageInDays < 7:
                str_age = '%03dD' % ageInDays
            return str_age

        def ageInDays_calculateExplicitly(d_DICOM : dict) -> int:
            """
            Explicitly calculate the PatientAge using the StudyDate and PatientBirthDate
            tag values.

            If unable to calculate return an age string of '999Y' otherwise, return the
            age in DICOM-y format i.e. '005Y' or '012M' etc.
            """
            d_pacsData  : dict  = d_DICOM['d_dicomSimple']
            ageInDays   : int   = 0
            str_age     : str   = '999Y'
            if len(d_pacsData['StudyDate']) and len(d_pacsData['PatientBirthDate']):
                try:
                    date_study = datetime.strptime(d_pacsData['StudyDate'], "%Y-%m-%d")
                except:
                    date_study = datetime.strptime(d_pacsData['StudyDate'], "%Y%m%d")
                try:
                    date_birth  = datetime.strptime(d_pacsData['PatientBirthDate'], "%Y-%m-%d")
                except:
                    date_birth  = datetime.strptime(d_pacsData['PatientBirthDate'], "%Y%m%d")
                ageInDays   = abs((date_study - date_birth).days)
                # str_age     = age_daysToDMY(ageInDays)
            return ageInDays

        def AgeInDays_set(d_DICOM : dict, ageInDays : int):
            """Add a trailing string of age in days (explicitly calculated)
            to the DICOM PatientAge field.

            Args:
                d_DICOM (dict): Existing DICOM tag structure
                ageInDays (int): The explicit age in days calculation

            Returns:
                Modifies the d_DICOM structure in place
            """
            if 'PatientAge' not in d_DICOM['d_dicomSimple']:
                d_DICOM['d_dicomSimple']['PatientAge']  = age_daysToDMY(ageInDays)
            str_DICOMage    : str   = d_DICOM['d_dicomSimple']['PatientAge']
            d_DICOM['d_dicomSimple']['PatientAge'] = '%s-%06dd' % (str_DICOMage, ageInDays)

        def ageTag_processAndUpdate(    str_tag     : str,
                                        d_DICOM     : dict,
                                        l_tags      : list,
                                        l_tagsToSub : list):
            """ Process age related tags in the DICOM structure --
                either the PatientAge and AgeInDays tags. If not currently
                present in the d_DICOM structure, then explicitly add to
                dictionary and tag lists.

            Args:
                str_tag (str):  tag to process
                d_DICOM (dict): DICOM header structure
                l_tags (list): original tag list
                l_tagsToSub (list): tag list found in DICOM header

            Returns:
                Modifies the d_DICOM and l_tagsToSub in place if necessary
            """
            if any(str_tag in string for string in l_tags) and not \
               any(str_tag in string for string in l_tagsToSub):
                ageInDays   : int   = ageInDays_calculateExplicitly(d_DICOM)
                if str_tag == 'PatientAge':
                    d_DICOM['d_dicomSimple'][str_tag] = age_daysToDMY(ageInDays)
                if str_tag == 'AgeInDays':
                    d_DICOM['d_dicomSimple'][str_tag] = '%06dd' % ageInDays
                l_tagsToSub.append(str_tag)

        b_tagsFound     : bool  = False
        str_replace     : str   = ''        # The lookup/processed tag value
        l_tags          : list  = []        # The input string split by '%'
        l_tagsToSub     : list  = []        # Remove any noise etc from each tag
        func            : str   = ''        # the function to apply
        tag             : str   = ''        # the tag in the funcTag combo

        if '%' in astr:
            l_tags          = astr.split('%')[1:]
            # Find which tags (mangled) in string match actual tags
            l_tagsToSub     = [i for i in d_DICOM['l_tagRaw'] \
                                if any(i in b for b in l_tags)]

            # Explicitly check the PatientAge and the AgeInDays tags
            for tag in ['PatientAge', 'AgeInDays']:
                ageTag_processAndUpdate(tag,  d_DICOM, l_tags, l_tagsToSub)

            # Make sure lists are in same sort order
            l_tagsToSubSort =  sorted(
                l_tagsToSub,
                key = lambda x: [i for i, s in enumerate(l_tags) if x in s][0]
            )
            for tag, func in zip(l_tagsToSubSort, l_tags):
                b_tagsFound     = True
                str_replace     = str(d_DICOM['d_dicomSimple'][tag])
                if 'md5'    in func: astr, str_replace   = md5_process(func, str_replace)
                if 'strmsk' in func: astr, str_replace   = strmsk_process(func, str_replace)
                if 'nospc'  in func: astr, str_replace   = nospc_process(func, str_replace)
                if 'name'   in func: astr, str_replace   = name_process(func, str_replace)
                if 'pad'    in func: astr, str_replace   = pad_process(func, str_replace)
                astr  = astr.replace('%' + tag, str_replace)

        return {
            'status':       True,
            'b_tagsFound':  b_tagsFound,
            'str_result':   astr
        }




"""
Redis database schema definition and helpers for "Really Efficient" mode.
"""
import os
from typing import TypedDict
import functools

try:
    import redis.asyncio as redis
    has_redis = True
except ImportError:
    has_redis = False


def needs_redis(fn):
    """A decorator for functions which require Redis."""
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        if not has_redis:
            raise RuntimeError('redis must be installed to call this function. Please run `pip install redis`')
        return fn(*args, **kwargs)
    return wrapper



def redisKeyOf(str_SeriesInstanceUID: str) -> str:
    """Produce the redis key for a series."""
    return f'series:{str_SeriesInstanceUID}'


class ReData(TypedDict):
    """
    Series pull progress information.
    """
    NumberOfSeriesRelatedInstances: int
    """Number of DICOM files in the series, reported by px-find (findscu)"""
    fileCounter: int
    """Number of files received by storescp, incremented by px-recount"""
    lastUpdate: str
    """Timestamp in ISO8901 format."""


@needs_redis
async def getRedisClient() -> 'redis.Redis':
    """
    Create the redis client.
    """
    if (url := os.getenv('PYPX_REDIS_URL', None)) is None:
        return redis.Redis(decode_responses=True)
    return await redis.from_url(url)



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()

setup(
        name                =   'pypx',
        version             =   '3.11.0rc1',
        description         =   'PACS/ChRIS core tools and utils',
        long_description    =   readme(),
        python_requires     =   ">= 3.8",
        url                 =   'http://github.com/FNNDSC/pypx',
        author              =   'FNNDSC Developers',
        author_email        =   'dev@babyMRI.org',
        license             =   'MIT',
        packages            =   ['pypx'],
        install_requires=[
            'pudb',
            'terminaltables',
            'py-dateutil',
            'pydicom',
            'pfmisc',
            'dask',
            'retry',
            'psutil',
            'python-swiftclient',
            'pfstate',
            'webob',
            'python-chrisclient',
        ],
        test_suite          =   'nose.collector',
        tests_require       =   ['nose'],
        scripts             =   [
            'storescp.sh',
            'bin/pfstorage',
            'bin/px-do',
            'bin/px-echo',
            'bin/px-find',
            'bin/px-listen',
            'bin/px-move',
            'bin/px-push',
            'bin/px-register',
            'bin/px-repack',
            'bin/px-report',
            'bin/px-smdb',
            'bin/px-status'
        ],
        zip_safe            =   False,
        extras_require={
            "none": [],
            "re": [
                'redis[hiredis]~=4.6',
                'px-recount==3.11.0-rc1'
            ],
        },
        )



import  subprocess, re, collections, codecs
import  asyncio
import  pfmisc
import  pudb
from    pfmisc._colors      import  Colors

# Credit to Anatoly Techtonik
# https://stackoverflow.com/questions/606191/convert-bytes-to-a-string/27527728#27527728
def slashescape(err):
    """ codecs error handler. err is UnicodeDecode instance. return
    a tuple with a replacement for the unencodable part of the input
    and a position where encoding should continue"""
    #print err, dir(err), err.start, err.end, err.object[:err.start]
    thebyte = err.object[err.start:err.end]
    repl = u'\\x'+hex(ord(thebyte))[2:]
    return (repl, err.end)

codecs.register_error('slashescape', slashescape)


class Base():
    """
    A Base class for pypx.

    This class is somewhat abstract/virtual in as much as it
    provides a substrate for derived classes and in and of
    itself does very little.

    """

    def defaults_init(self, d_arg):
        """
        Somewhat convoluted initialization that mirrors/duplicates
        the default initialization when called from CLI
        """

        for k,v in d_arg.items():
            if k in self.arg:
                setattr(self, k, self.arg[k])
            else:
                setattr(self, k, v)

    def __init__(self, arg):
        """
        Initialize some core self variables common to all derived classes.
        """

        self.arg = arg

        # For lint sanity, the following are explicitly set
        # (note that they are implicitly created in the call
        # to defaults_init):
        self.verbosity      = 1
        self.aec            = ''
        self.aet            = ''
        self.serverIP       = ''
        self.serverPort     = ''

        self.defaults_init(
            {
                'aet':          'CHRIS-AET',
                'aec':          'CHRIS-AEC',
                'serverIP':     '127.0.0.1',
                'serverPort':   '4242',
                'findscu':      '/usr/bin/findscu',
                'movescu':      '/usr/bin/movescu',
                'storescp':     '/usr/bin/storescp',
                'echoscu':      '/usr/bin/echoscu',
                'colorize':     '',
                'printReport':  '',
                'verbosity':    1,
                'retrieve':     False,
                'move':         False
            }
        )

        self.response = {
            'status': 'error',
            'data': {}
        }

        self.dp = pfmisc.debug(
                    verbosity   = self.verbosity,
                    within      = 'Base',
                    syslog      = False
        )

    def systemlevel_run(self, opt, d_params):
        """
        Run the system command, based on the passed parameter dictionary
        """
        b_commandGen    = False
        str_cmd         = ''
        d_ret           = {}

        # A function to execute to generate commands
        f_commandGen    = None
        for k,v in d_params.items():
            if k == 'f_commandGen':
                f_commandGen    = v
                b_commandGen    = True
            else:
                opt[k]  = v

        if b_commandGen:
            str_cmd         = f_commandGen(opt)
            self.dp.qprint("\n%s" % str_cmd, level = 5, type = 'status')
            raw_response    = subprocess.run(
                                str_cmd,
                                stdout  = subprocess.PIPE,
                                stderr  = subprocess.STDOUT,
                                shell   = True
            )
            d_ret   = self.formatResponse(raw_response)

        return d_ret

    async def systemlevel_runasync(self, opt, d_params):
        """
        Run the system command asynchronously, based on the passed parameter
        dictionary. Avoid using if large amounts of stdout data is generated!
        Output is assumed to be UTF-8 text.
        """
        b_commandGen    = False
        str_cmd         = ''
        d_ret           = {}

        # A function to execute to generate commands
        f_commandGen    = None
        for k,v in d_params.items():
            if k == 'f_commandGen':
                f_commandGen    = v
                b_commandGen    = True
            else:
                opt[k]  = v

        if b_commandGen:
            str_cmd         = f_commandGen(opt)
            self.dp.qprint("\n%s" % str_cmd, level = 5, type = 'status')
            raw_response    = await asyncio.create_subprocess_shell(
                                str_cmd,
                                stdout=asyncio.subprocess.PIPE,
                                stderr=asyncio.subprocess.STDOUT
                            )
            await raw_response.wait()

            stdout, stderr = await raw_response.communicate()

            d_ret   = self.__formatResponseHelper(
                stdout, str_cmd, raw_response.returncode
            )

        return d_ret

    def commandSuffix(self):
        # required parameters
        command_suffix =    ' -aec '    + self.aec
        command_suffix +=   ' -aet '    + self.aet
        command_suffix +=   ' '         + self.serverIP
        command_suffix +=   ' '         + self.serverPort

        return command_suffix

    def handle(self, raw_response):
        std = raw_response.stdout.decode('ascii')
        response = {
            'status': 'success',
            'data': '',
            'command': raw_response.args
        }
        if std != '':
            response['status'] = 'error'
            response['data'] = std

        return response

    def checkResponse(self, response):
        std_split = response.split('\n')
        info_count = 0
        error_count = 0
        for line in std_split:
            if line.startswith('I: '):
                info_count += 1
            elif line.startswith('E: ') or 'error' in line.lower():
                error_count += 1

        status = 'error'
        if error_count == 0:
            status = 'success'

        return status

    def parseResponse(self, response):
        data = []

        uid = 0
        std_split = response.split('\n')

        for line in std_split:
            if line.startswith('I: ---------------------------'):
                data.append({})
                data[-1]['uid'] = {}
                data[-1]['uid']['tag'] = 0
                data[-1]['uid']['value'] = uid
                data[-1]['uid']['label'] = 'uid'
                uid += 1

            elif line.startswith('I: '):
                lineSplit = line.split()
                if len(lineSplit) >= 8 and re.search('\((.*?)\)', lineSplit[1]) != None:
                    # extract DICOM tag
                    tag = re.search('\((.*?)\)', lineSplit[1]).group(0)[1:-1].strip().replace('\x00', '')

                    # extract value
                    value = re.search('\[(.*?)\]', line)
                    if value != None:
                        value = value.group(0)[1:-1].strip().replace('\x00', '')
                    else:
                        value = 'no value provided for %s' % tag

                    # extract label
                    label = lineSplit[-1].strip()

                    data[-1][label] = {}
                    data[-1][label]['tag'] = tag
                    data[-1][label]['value'] = value
                    data[-1][label]['label'] = label
                else:
                    # Only append the line output for the echo command
                    if type(self).__name__ == 'Echo': data.append(line)

        return data

    def formatResponse(self, raw_response):
        return self.__formatResponseHelper(
            raw_response.stdout,
            raw_response.args,
            raw_response.returncode
        )

    def __formatResponseHelper(self, stdout, args, returncode):
        std = stdout.decode('utf-8', 'slashescape')
        response = {
            'status':       'success',
            'data':         '',
            'command':      args,
            'returncode':   returncode
        }

        status = self.checkResponse(std)

        if status == 'error':
            response['status']  = 'error'
            response['data']    = std
        else:
            response['status']  = 'success'
            response['data']    = self.parseResponse(std)

        return response